import numpy as np
from ase.units import Bohr
from ase.utils.timing import timer
from gpaw.directmin.tools import loewdin_lcao, gramschmidt_lcao

from gpaw.lfc import BasisFunctions
from gpaw.utilities import unpack
from gpaw.utilities.tools import tri2full
# from gpaw import debug
# from gpaw.lcao.overlap import NewTwoCenterIntegrals as NewTCI
from gpaw.lcao.tci import TCIExpansions
from gpaw.utilities.blas import mmm, gemmdot
from gpaw.wavefunctions.base import WaveFunctions
from gpaw.lcao.atomic_correction import (DenseAtomicCorrection,
                                         SparseAtomicCorrection)
from gpaw.wavefunctions.mode import Mode
from gpaw.directmin.etdm import ETDM


class LCAO(Mode):
    name = 'lcao'

    def __init__(self, atomic_correction=None, interpolation=3,
                 force_complex_dtype=False):
        self.atomic_correction = atomic_correction
        self.interpolation = interpolation
        Mode.__init__(self, force_complex_dtype)

    def __call__(self, *args, **kwargs):
        return LCAOWaveFunctions(*args,
                                 atomic_correction=self.atomic_correction,
                                 **kwargs)

    def __repr__(self):
        return 'LCAO({})'.format(self.todict())

    def todict(self):
        dct = Mode.todict(self)
        dct['interpolation'] = self.interpolation
        return dct


def update_phases(C_unM, q_u, ibzk_qc, spos_ac, oldspos_ac, setups, Mstart):
    """Complex-rotate coefficients compensating discontinuous phase shift.

    This changes the coefficients to counteract the phase discontinuity
    of overlaps when atoms move across a cell boundary."""

    # We don't want to apply any phase shift unless we crossed a cell
    # boundary.  So we round the shift to either 0 or 1.
    #
    # Example: spos_ac goes from 0.01 to 0.99 -- this rounds to 1 and
    # we apply the phase.  If someone moves an atom by half a cell
    # without crossing a boundary, then we are out of luck.  But they
    # should have reinitialized from LCAO anyway.
    phase_qa = np.exp(2j * np.pi *
                      np.dot(ibzk_qc, (spos_ac - oldspos_ac).T.round()))

    for q, C_nM in zip(q_u, C_unM):
        if C_nM is None:
            continue
        for a in range(len(spos_ac)):
            M1 = setups.M_a[a] - Mstart
            M2 = M1 + setups[a].nao
            M1 = max(0, M1)
            C_nM[:, M1:M2] *= phase_qa[q, a]  # (may truncate M2)


# replace by class to make data structure perhaps a bit less confusing
def get_r_and_offsets(nl, spos_ac, cell_cv):
    r_and_offset_aao = {}

    def add(a1, a2, R_c, offset):
        if not (a1, a2) in r_and_offset_aao:
            r_and_offset_aao[(a1, a2)] = []
        r_and_offset_aao[(a1, a2)].append((R_c, offset))

    for a1, spos1_c in enumerate(spos_ac):
        a2_a, offsets = nl.get_neighbors(a1)
        for a2, offset in zip(a2_a, offsets):
            spos2_c = spos_ac[a2] + offset

            R_c = np.dot(spos2_c - spos1_c, cell_cv)
            add(a1, a2, R_c, offset)
            if a1 != a2 or offset.any():
                add(a2, a1, -R_c, -offset)

    return r_and_offset_aao


class LCAOWaveFunctions(WaveFunctions):
    mode = 'lcao'

    def __init__(self, ksl, gd, nvalence, setups, bd,
                 dtype, world, kd, kptband_comm, timer,
                 atomic_correction=None, collinear=True):
        WaveFunctions.__init__(self, gd, nvalence, setups, bd,
                               dtype, collinear, world, kd,
                               kptband_comm, timer)
        self.ksl = ksl
        self.S_qMM = None
        self.T_qMM = None
        self.P_aqMi = None
        self.debug_tci = False

        if atomic_correction is None:
            atomic_correction = 'sparse' if ksl.using_blacs else 'dense'

        if atomic_correction == 'sparse':
            self.atomic_correction_cls = SparseAtomicCorrection
        else:
            assert atomic_correction == 'dense'
            self.atomic_correction_cls = DenseAtomicCorrection

        # self.tci = NewTCI(gd.cell_cv, gd.pbc_c, setups, kd.ibzk_qc, kd.gamma)
        with self.timer('TCI: Evaluate splines'):
            self.tciexpansions = TCIExpansions.new_from_setups(setups)

        self.basis_functions = BasisFunctions(gd,
                                              [setup.basis_functions_J
                                               for setup in setups],
                                              kd,
                                              dtype=dtype,
                                              cut=True)

        self.coefficients_read_from_file = False
        self.set_orthonormalized(False)

    def set_orthonormalized(self, flag):
        self.orthonormalized = flag

    @timer('Orthonormalize')
    def orthonormalize(self, kpt=None, type='gramschmidt'):
        assert type == 'gramschmidt' or type == 'loewdin'
        if kpt is None:
            for kpt in self.kpt_u:
                self.orthonormalize(kpt)
            self.orthonormalized = True
            return
        if type == 'loewdin':
            kpt.C_nM[:] = loewdin_lcao(kpt.C_nM, kpt.S_MM.conj())
        elif type == 'gramschmidt':
            kpt.C_nM[:] = gramschmidt_lcao(kpt.C_nM, kpt.S_MM.conj())

    def empty(self, n=(), global_array=False, realspace=False):
        if realspace:
            return self.gd.empty(n, self.dtype, global_array)
        else:
            if isinstance(n, int):
                n = (n,)
            nao = self.setups.nao
            return np.empty(n + (nao,), self.dtype)

    def __str__(self):
        s = 'Wave functions: LCAO\n'
        s += '  Diagonalizer: %s\n' % self.ksl.get_description()
        s += ('  Atomic Correction: %s\n'
              % self.atomic_correction_cls.description)
        s += '  Data-type: %s\n' % self.dtype.__name__
        return s

    def set_eigensolver(self, eigensolver):
        WaveFunctions.set_eigensolver(self, eigensolver)
        if eigensolver:
            if isinstance(eigensolver, ETDM):
                eigensolver.initialize(self.gd, self.dtype, self.bd.nbands,
                                       self.kd.nibzkpts, self.setups.nao,
                                       self.ksl.using_blacs,
                                       self.bd.comm.size, self.kpt_u)
            else:
                eigensolver.initialize(self.gd, self.dtype, self.setups.nao,
                                       self.ksl)

    def set_positions(self, spos_ac, atom_partition=None, move_wfs=False):
        oldspos_ac = self.spos_ac
        with self.timer('Basic WFS set positions'):
            WaveFunctions.set_positions(self, spos_ac, atom_partition)

        with self.timer('Basis functions set positions'):
            self.basis_functions.set_positions(spos_ac)

        if self.ksl is not None:
            self.basis_functions.set_matrix_distribution(self.ksl.Mstart,
                                                         self.ksl.Mstop)

        Mstop = self.ksl.Mstop
        Mstart = self.ksl.Mstart

        # if self.ksl.using_blacs:  # XXX
        #     S and T have been distributed to a layout with blacs, so
        #     discard them to force reallocation from scratch.
        #
        #     TODO: evaluate S and T when they *are* distributed, thus saving
        #     memory and avoiding this problem
        for kpt in self.kpt_u:
            kpt.S_MM = None
            kpt.T_MM = None

        # Free memory in case of old matrices:
        self.S_qMM = self.T_qMM = self.P_aqMi = None

        if self.dtype == complex and oldspos_ac is not None:
            update_phases([kpt.C_nM for kpt in self.kpt_u],
                          [kpt.q for kpt in self.kpt_u],
                          self.kd.ibzk_qc, spos_ac, oldspos_ac,
                          self.setups, Mstart)

        self.timer.start('mktci')
        manytci = self.tciexpansions.get_manytci_calculator(
            self.setups, self.gd, spos_ac, self.kd.ibzk_qc, self.dtype,
            self.timer)
        self.timer.stop('mktci')
        self.manytci = manytci
        self.newtci = manytci.tci

        my_atom_indices = self.basis_functions.my_atom_indices
        self.timer.start('ST tci')
        newS_qMM, newT_qMM = manytci.O_qMM_T_qMM(self.gd.comm,
                                                 Mstart, Mstop,
                                                 self.ksl.using_blacs)
        self.timer.stop('ST tci')
        self.timer.start('P tci')
        P_qIM = manytci.P_qIM(my_atom_indices)
        self.timer.stop('P tci')
        self.P_aqMi = manytci.P_aqMi(my_atom_indices)
        self.P_qIM = P_qIM  # XXX atomic correction

        self.atomic_correction = self.atomic_correction_cls.new_from_wfs(self)

        # TODO
        #   OK complex/conj, periodic images
        #   OK scalapack
        #   derivatives/forces
        #   sparse
        #   use symmetry/conj tricks to reduce calculations
        #   enable caching of spherical harmonics

        self.atomic_correction.add_overlap_correction(newS_qMM)
        self.allocate_arrays_for_projections(my_atom_indices)

        newS_qMM = self.ksl.distribute_overlap_matrix(newS_qMM, root=-1)
        newT_qMM = self.ksl.distribute_overlap_matrix(newT_qMM, root=-1)

        self.positions_set = True

        for kpt in self.kpt_u:
            q = kpt.q
            kpt.S_MM = newS_qMM[q]
            kpt.T_MM = newT_qMM[q]
        self.S_qMM = newS_qMM
        self.T_qMM = newT_qMM

        # Elpa wants to reuse the decomposed form of S_qMM.
        # We need to keep track of the existence of that object here,
        # since this is where we change S_qMM.  Hence, expect this to
        # become arrays after the first diagonalization:
        self.decomposed_S_qMM = [None] * len(self.S_qMM)
        self.set_orthonormalized(False)

    def initialize(self, density, hamiltonian, spos_ac):
        # Note: The above line exists also in set_positions.
        # This is guaranteed to be correct, but we can probably remove one.
        # Of course no human can understand the initialization process,
        # so this will be some other day.
        self.timer.start('LCAO WFS Initialize')
        if density.nt_sG is None:
            if self.kpt_u[0].f_n is None or self.kpt_u[0].C_nM is None:
                density.initialize_from_atomic_densities(self.basis_functions)
            else:
                # We have the info we need for a density matrix, so initialize
                # from that instead of from scratch.  This will be the case
                # after set_positions() during a relaxation
                density.initialize_from_wavefunctions(self)
            # Initialize GLLB-potential from basis function orbitals
            if hamiltonian.xc.type == 'GLLB':
                hamiltonian.xc.initialize_from_atomic_orbitals(
                    self.basis_functions)

        else:
            # After a restart, nt_sg doesn't exist yet, so we'll have to
            # make sure it does.  Of course, this should have been taken care
            # of already by this time, so we should improve the code elsewhere
            density.calculate_normalized_charges_and_mix()

        hamiltonian.update(density)
        self.timer.stop('LCAO WFS Initialize')

        return 0, 0

    def initialize_wave_functions_from_lcao(self):
        """Fill the calc.wfs.kpt_[u].psit_nG arrays with useful data.

        Normally psit_nG is NOT used in lcao mode, but some extensions
        (like ase.dft.wannier) want to have it.
        This code is adapted from fd.py / initialize_from_lcao_coefficients()
        and fills psit_nG with data constructed from the current lcao
        coefficients (kpt.C_nM).

        (This may or may not work in band-parallel case!)
        """
        from gpaw.wavefunctions.arrays import UniformGridWaveFunctions
        bfs = self.basis_functions
        for kpt in self.kpt_u:
            kpt.psit = UniformGridWaveFunctions(
                self.bd.nbands, self.gd, self.dtype, kpt=kpt.q, dist=None,
                spin=kpt.s, collinear=True)
            kpt.psit_nG[:] = 0.0
            bfs.lcao_to_grid(kpt.C_nM, kpt.psit_nG[:self.bd.mynbands], kpt.q)

    def initialize_wave_functions_from_restart_file(self):
        """Dummy function to ensure compatibility to fd mode"""
        self.initialize_wave_functions_from_lcao()

    def add_orbital_density(self, nt_G, kpt, n):
        rank, q = self.kd.get_rank_and_index(kpt.k)
        u = q * self.nspins + kpt.s
        assert rank == self.kd.comm.rank
        assert self.kpt_u[u] is kpt
        psit_G = self._get_wave_function_array(u, n, realspace=True)
        self.add_realspace_orbital_to_density(nt_G, psit_G)

    def calculate_density_matrix(self, f_n, C_nM, rho_MM=None):
        self.timer.start('Calculate density matrix')
        rho_MM = self.ksl.calculate_density_matrix(f_n, C_nM, rho_MM)
        self.timer.stop('Calculate density matrix')
        return rho_MM

        if 1:
            # XXX Should not conjugate, but call gemm(..., 'c')
            # Although that requires knowing C_Mn and not C_nM.
            # that also conforms better to the usual conventions in literature
            Cf_Mn = C_nM.T.conj() * f_n
            self.timer.start('gemm')
            mmm(1.0, Cf_Mn, 'N', C_nM, 'N', 0.0, rho_MM)
            self.timer.stop('gemm')
            self.timer.start('band comm sum')
            self.bd.comm.sum(rho_MM)
            self.timer.stop('band comm sum')
        else:
            # Alternative suggestion. Might be faster. Someone should test this
            from gpaw.utilities.blas import r2k
            C_Mn = C_nM.T.copy()
            r2k(0.5, C_Mn, f_n * C_Mn, 0.0, rho_MM)
            tri2full(rho_MM)

    def calculate_atomic_density_matrices_with_occupation(self, D_asp, f_un):
        # ac = self.atomic_correction
        # if ac.implements_distributed_projections():
        #     D2_asp = ac.redistribute(self, D_asp, type='asp', op='forth')
        #     WaveFunctions.calculate_atomic_density_matrices_with_occupation(
        #         self, D2_asp, f_un)
        #     D3_asp = ac.redistribute(self, D2_asp, type='asp', op='back')
        #     for a in D_asp:
        #         D_asp[a][:] = D3_asp[a]
        # else:
        WaveFunctions.calculate_atomic_density_matrices_with_occupation(
            self, D_asp, f_un)

    def calculate_density_matrix_delta(self, d_nn, C_nM, rho_MM=None):
        self.timer.start('Calculate density matrix')
        rho_MM = self.ksl.calculate_density_matrix_delta(d_nn, C_nM, rho_MM)
        self.timer.stop('Calculate density matrix')
        return rho_MM

    def add_to_density_from_k_point_with_occupation(self, nt_sG, kpt, f_n):
        """Add contribution to pseudo electron-density. Do not use the standard
        occupation numbers, but ones given with argument f_n."""
        # Custom occupations are used in calculation of response potential
        # with GLLB-potential
        if kpt.rho_MM is None:
            rho_MM = self.calculate_density_matrix(f_n, kpt.C_nM)
            if hasattr(kpt, 'c_on'):
                assert self.bd.comm.size == 1
                d_nn = np.zeros((self.bd.mynbands, self.bd.mynbands),
                                dtype=kpt.C_nM.dtype)
                for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                    assert abs(c_n.imag).max() < 1e-14
                    d_nn += ne * np.outer(c_n.conj(), c_n).real
                rho_MM += self.calculate_density_matrix_delta(d_nn, kpt.C_nM)
        else:
            rho_MM = kpt.rho_MM
        self.timer.start('Construct density')
        self.basis_functions.construct_density(rho_MM, nt_sG[kpt.s], kpt.q)
        self.timer.stop('Construct density')

    def add_to_kinetic_density_from_k_point(self, taut_G, kpt):
        raise NotImplementedError('Kinetic density calculation for LCAO '
                                  'wavefunctions is not implemented.')

    def calculate_forces(self, hamiltonian, F_av):
        self.timer.start('LCAO forces')

        Fref_av = np.zeros_like(F_av)
        self.forcecalc = LCAOforces(self.ksl, self.dtype, self.gd,
                                    self.bd, self.kd, self.kpt_u, self.nspins,
                                    self.basis_functions, self.newtci,
                                    self.P_aqMi, self.setups,
                                    self.manytci, hamiltonian,
                                    self, self.spos_ac,
                                    self.timer, Fref_av)

        F_av[:, :] = self.forcecalc.get_forces_sum_GS()
        self.timer.stop('LCAO forces')

    def _get_wave_function_array(self, u, n, realspace=True, periodic=False):
        # XXX Taking kpt is better than taking u
        kpt = self.kpt_u[u]
        C_M = kpt.C_nM[n]

        if realspace:
            psit_G = self.gd.zeros(dtype=self.dtype)
            self.basis_functions.lcao_to_grid(C_M, psit_G, kpt.q)
            if periodic and self.dtype == complex:
                k_c = self.kd.ibzk_kc[kpt.k]
                return self.gd.plane_wave(-k_c) * psit_G
            return psit_G
        else:
            return C_M

    def write(self, writer, write_wave_functions=False):
        WaveFunctions.write(self, writer)
        if write_wave_functions:
            self.write_wave_functions(writer)

    def write_wave_functions(self, writer):
        writer.add_array(
            'coefficients',
            (self.nspins, self.kd.nibzkpts, self.bd.nbands, self.setups.nao),
            dtype=self.dtype)
        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                C_nM = self.collect_array('C_nM', k, s)
                writer.fill(C_nM * Bohr**-1.5)

    def read(self, reader):
        WaveFunctions.read(self, reader)
        r = reader.wave_functions
        if 'coefficients' in r:
            self.read_wave_functions(r)

    def read_wave_functions(self, reader):
        for kpt in self.kpt_u:
            C_nM = reader.proxy('coefficients', kpt.s, kpt.k)
            kpt.C_nM = self.bd.empty(self.setups.nao, dtype=self.dtype)
            for myn, C_M in enumerate(kpt.C_nM):
                n = self.bd.global_index(myn)
                # XXX number of bands could have been rounded up!
                if n >= len(C_nM):
                    break
                C_M[:] = C_nM[n] * Bohr**1.5

        self.coefficients_read_from_file = True

    def estimate_memory(self, mem):
        nq = len(self.kd.ibzk_qc)
        nao = self.setups.nao
        ni_total = sum([setup.ni for setup in self.setups])
        itemsize = mem.itemsize[self.dtype]
        mem.subnode('C [qnM]', nq * self.bd.mynbands * nao * itemsize)
        nM1, nM2 = self.ksl.get_overlap_matrix_shape()
        mem.subnode('S, T [2 x qmm]', 2 * nq * nM1 * nM2 * itemsize)
        mem.subnode('P [aqMi]', nq * nao * ni_total // self.gd.comm.size)
        # self.tci.estimate_memory(mem.subnode('TCI'))
        self.basis_functions.estimate_memory(mem.subnode('BasisFunctions'))
        self.eigensolver.estimate_memory(mem.subnode('Eigensolver'),
                                         self.dtype)


class LCAOforces:

    def __init__(self, ksl, dtype, gd, bd, kd, kpt_u, nspins, bfs, newtci,
                 P_aqMi, setups, manytci, hamiltonian, wfs, spos_ac,
                 timer, Fref_av):
        """ Object which calculates LCAO forces """

        self.ksl = ksl
        self.nao = ksl.nao
        self.mynao = ksl.mynao
        self.dtype = dtype
        self.newtci = newtci
        self.manytci = manytci
        self.P_aqMi = P_aqMi
        self.gd = gd
        self.bd = bd
        self.kd = kd
        self.kpt_u = kpt_u
        self.nspins = nspins
        self.bfs = bfs
        self.spos_ac = spos_ac
        self.Mstart = ksl.Mstart
        self.Mstop = ksl.Mstop
        self.setups = setups
        self.hamiltonian = hamiltonian
        self.wfs = wfs
        self.timer = timer
        self.Fref_av = Fref_av
        self.my_atom_indices = bfs.my_atom_indices
        self.atom_indices = bfs.atom_indices
        self.dH_asp = hamiltonian.dH_asp

        from gpaw.kohnsham_layouts import BlacsOrbitalLayouts
        self.isblacs = isinstance(self.ksl, BlacsOrbitalLayouts)

        if not self.isblacs:
            self.timer.start('TCI derivative')
            self.dThetadR_qvMM, self.dTdR_qvMM = self.manytci.O_qMM_T_qMM(
                self.gd.comm, self.Mstart, self.Mstop, False, derivative=True)
            self.dPdR_aqvMi = self.manytci.P_aqMi(self.bfs.my_atom_indices,
                                                  derivative=True)

            self.gd.comm.sum(self.dThetadR_qvMM)
            self.gd.comm.sum(self.dTdR_qvMM)
            self.timer.stop('TCI derivative')
            self.rhoT_uMM, self.ET_uMM = self.get_den_mat_and_E()

    def get_forces_sum_GS(self):
        """ This function calculates ground state forces in LCAO mode """
        if not self.isblacs:
            F_av = np.zeros_like(self.Fref_av)
            Fkin_av = self.get_kinetic_term()
            Fpot_av = self.get_pot_term()
            Ftheta_av = self.get_den_mat_term()
            Frho_av = self.get_den_mat_paw_term()
            Fatom_av = self.get_atomic_density_term()
            F_av += Fkin_av + Fpot_av + Ftheta_av + Frho_av + Fatom_av
        else:
            F_av = np.zeros_like(self.Fref_av)
            Fpot_av = self.get_pot_term_blacs()
            Fkin_av, Ftheta_av = self.get_kin_and_den_term_blacs()
            Fatom_av, Frho_av = self.get_at_den_and_den_paw_blacs()

            F_av += Fkin_av + Fpot_av + Ftheta_av + Frho_av + Fatom_av

        self.timer.start('Wait for sum')
        self.ksl.orbital_comm.sum(F_av)
        if self.bd.comm.rank == 0:
            self.kd.comm.sum(F_av, 0)
        self.timer.stop('Wait for sum')

        return F_av

    def _slices(self, indices):
        for a in indices:
            M1 = self.bfs.M_a[a] - self.Mstart
            M2 = M1 + self.setups[a].nao
            if M2 > 0:
                yield a, max(0, M1), M2

    def slices(self):
        return self._slices(self.atom_indices)

    def my_slices(self):
        return self._slices(self.my_atom_indices)

    def get_den_mat_and_E(self):
        #
        #         -----                    -----
        #          \    -1                  \    *
        # E      =  )  S     H    rho     =  )  c     eps  f  c
        #  mu nu   /    mu x  x z    z nu   /    n mu    n  n  n nu
        #         -----                    -----
        #          x z                       n
        #
        # We use the transpose of that matrix.  The first form is used
        # if rho is given, otherwise the coefficients are used.
        self.timer.start('Initial')
        if self.kpt_u[0].rho_MM is None:
            rhoT_uMM = []
            ET_uMM = []
            self.timer.start('Get density matrix')
            for kpt in self.kpt_u:
                rhoT_MM = self.ksl.get_transposed_density_matrix(kpt.f_n,
                                                                 kpt.C_nM)
                rhoT_uMM.append(rhoT_MM)
                ET_MM = self.ksl.get_transposed_density_matrix(kpt.f_n *
                                                               kpt.eps_n,
                                                               kpt.C_nM)
                ET_uMM.append(ET_MM)
                if hasattr(kpt, 'c_on'):
                    # XXX does this work with BLACS/non-BLACS/etc.?
                    assert self.bd.comm.size == 1
                    d_nn = np.zeros((self.bd.mynbands, self.bd.mynbands),
                                    dtype=kpt.C_nM.dtype)
                    for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                        d_nn += ne * np.outer(c_n.conj(), c_n)
                    rhoT_MM += self.ksl.get_transposed_density_matrix_delta(
                        d_nn, kpt.C_nM)
                    ET_MM += self.ksl.get_transposed_density_matrix_delta(
                        d_nn * kpt.eps_n, kpt.C_nM)
            self.timer.stop('Get density matrix')
        else:
            rhoT_uMM = []
            ET_uMM = []
            for kpt in self.kpt_u:
                H_MM = self.wfs.eigensolver.calculate_hamiltonian_matrix(
                    self.hamiltonian, self.wfs, kpt)
                tri2full(H_MM)
                S_MM = kpt.S_MM.copy()
                tri2full(S_MM)
                ET_MM = np.linalg.solve(S_MM, gemmdot(H_MM,
                                                      kpt.rho_MM)).T.copy()
                del S_MM, H_MM
                rhoT_MM = kpt.rho_MM.T.copy()
                rhoT_uMM.append(rhoT_MM)
                ET_uMM.append(ET_MM)
        self.timer.stop('Initial')
        return rhoT_uMM, ET_uMM

    def get_kinetic_term(self):
        """Calculate Kinetic energy term in LCAO"""
        Fkin_av = np.zeros_like(self.Fref_av)
        self.timer.start('TCI derivative')
        # Kinetic energy contribution
        #
        #           ----- d T
        #  a         \       mu nu
        # F += 2 Re   )   -------- rho
        #            /    d R         nu mu
        #           -----    mu nu
        #        mu in a; nu
        #
        Fkin_av = np.zeros_like(Fkin_av)
        for u, kpt in enumerate(self.kpt_u):
            dEdTrhoT_vMM = (self.dTdR_qvMM[kpt.q] *
                            self.rhoT_uMM[u][np.newaxis]).real
            # XXX load distribution!
            for a, M1, M2 in self.my_slices():
                Fkin_av[a, :] += \
                    2.0 * dEdTrhoT_vMM[:, M1:M2].sum(-1).sum(-1)
        self.timer.stop('TCI derivative')

        return Fkin_av

    def get_den_mat_term(self):
        """Calculate density matrix term in LCAO"""
        Ftheta_av = np.zeros_like(self.Fref_av)
        # Density matrix contribution due to basis overlap
        #
        #            ----- d Theta
        #  a          \           mu nu
        # F  += -2 Re  )   ------------  E
        #             /        d R        nu mu
        #            -----        mu nu
        #         mu in a; nu
        #
        Ftheta_av = np.zeros_like(Ftheta_av)
        for u, kpt in enumerate(self.kpt_u):
            dThetadRE_vMM = (self.dThetadR_qvMM[kpt.q] *
                             self.ET_uMM[u][np.newaxis]).real
            for a, M1, M2 in self.my_slices():
                Ftheta_av[a, :] += \
                    -2.0 * dThetadRE_vMM[:, M1:M2].sum(-1).sum(-1)
        return Ftheta_av

    def get_pot_term(self):
        """Calculate potential term"""
        Fpot_av = np.zeros_like(self.Fref_av)
        # Potential contribution
        #
        #           -----      /  d Phi  (r)
        #  a         \        |        mu    ~
        # F += -2 Re  )       |   ---------- v (r)  Phi  (r) dr rho
        #            /        |     d R                nu          nu mu
        #           -----    /         a
        #        mu in a; nu
        #
        self.timer.start('Potential')
        vt_sG = self.hamiltonian.vt_sG
        Fpot_av = np.zeros_like(Fpot_av)
        for u, kpt in enumerate(self.kpt_u):
            vt_G = vt_sG[kpt.s]
            Fpot_av += self.bfs.calculate_force_contribution(vt_G,
                                                             self.rhoT_uMM[u],
                                                             kpt.q)
        self.timer.stop('Potential')

        return Fpot_av

    def get_den_mat_paw_term(self):
        """Calcualte PAW correction"""
        # TO DO: split this function into
        # _get_den_mat_paw_term (which calculate Frho_av) and
        # get_paw_correction (which calculate ZE_MM)
        # Density matrix contribution from PAW correction
        #
        #           -----                        -----
        #  a         \      a                     \     b
        # F +=  2 Re  )    Z      E        - 2 Re  )   Z      E
        #            /      mu nu  nu mu          /     mu nu  nu mu
        #           -----                        -----
        #           mu nu                    b; mu in a; nu
        #
        # with
        #                  b*
        #         -----  dP
        #   b      \       i mu    b   b
        #  Z     =  )   -------- dS   P
        #   mu nu  /     dR        ij  j nu
        #         -----    b mu
        #           ij
        #
        self.timer.start('Paw correction')
        Frho_av = np.zeros_like(self.Fref_av)
        for u, kpt in enumerate(self.kpt_u):
            work_MM = np.zeros((self.mynao, self.nao), self.dtype)
            ZE_MM = None
            for b in self.my_atom_indices:
                setup = self.setups[b]
                dO_ii = np.asarray(setup.dO_ii, self.dtype)
                dOP_iM = np.zeros((setup.ni, self.nao), self.dtype)
                mmm(1.0, dO_ii, 'N', self.P_aqMi[b][kpt.q], 'C', 0.0, dOP_iM)
                for v in range(3):
                    mmm(1.0,
                        self.dPdR_aqvMi[b][kpt.q][v][self.Mstart:self.Mstop],
                        'N',
                        dOP_iM, 'N',
                        0.0, work_MM)
                    ZE_MM = (work_MM * self.ET_uMM[u]).real
                    for a, M1, M2 in self.slices():
                        dE = 2 * ZE_MM[M1:M2].sum()
                        Frho_av[a, v] -= dE  # the "b; mu in a; nu" term
                        Frho_av[b, v] += dE  # the "mu nu" term
        self.timer.stop('Paw correction')
        return Frho_av

    def _get_den_mat_paw_term(self):
        # THIS doesn't work in parallel
        # Density matrix contribution from PAW correction
        #
        #           -----                        -----
        #  a         \      a                     \     b
        # F +=  2 Re  )    Z      E        - 2 Re  )   Z      E
        #            /      mu nu  nu mu          /     mu nu  nu mu
        #           -----                        -----
        #           mu nu                    b; mu in a; nu
        #
        # with
        #                  b*
        #         -----  dP
        #   b      \       i mu    b   b
        #  Z     =  )   -------- dS   P
        #   mu nu  /     dR        ij  j nu
        #         -----    b mu
        #           ij
        #
        Frho_av = np.zeros_like(self.Fref_av)
        self.timer.start('add paw correction')
        ZE_MM = self.get_paw_correction()
        for u, kpt in enumerate(self.kpt_u):
            for b in self.my_atom_indices:
                for v in range(3):
                    for a, M1, M2 in self.slices():
                        dE = 2 * ZE_MM[u, b, v, M1:M2].sum()
                        Frho_av[a, v] -= dE.real  # the "b; mu in a; nu" term
                        Frho_av[b, v] += dE.real  # the "mu nu" term
        self.timer.stop('add paw correction')
        return Frho_av

    def get_paw_correction(self):
        # THIS doesn't work in parallel
        #  <Phi_nu|pt_i>O_ii<dPt_i/dR|Phi_mu>
        self.timer.start('get paw correction')
        ZE_MM = np.zeros((len(self.kpt_u), len(self.my_atom_indices), 3,
                          self.mynao, self.nao), self.dtype)
        for u, kpt in enumerate(self.kpt_u):
            work_MM = np.zeros((self.mynao, self.nao), self.dtype)
            for b in self.my_atom_indices:
                setup = self.setups[b]
                dO_ii = np.asarray(setup.dO_ii, self.dtype)
                dOP_iM = np.zeros((setup.ni, self.nao), self.dtype)
                mmm(1.0, dO_ii, 'N', self.P_aqMi[b][kpt.q], 'C', 0.0, dOP_iM)
                for v in range(3):
                    mmm(1.0,
                        self.dPdR_aqvMi[b][kpt.q][v][self.Mstart:self.Mstop],
                        'N',
                        dOP_iM, 'N',
                        0.0, work_MM)
                    ZE_MM[u, b, v, :, :] = (work_MM * self.ET_uMM[u]).real
        self.timer.stop('get paw correction')
        return ZE_MM

    def get_atomic_density_term(self):
        Fatom_av = np.zeros_like(self.Fref_av)
        # Atomic density contribution
        #            -----                         -----
        #  a          \     a                       \     b
        # F  += -2 Re  )   A      rho       + 2 Re   )   A      rho
        #             /     mu nu    nu mu          /     mu nu    nu mu
        #            -----                         -----
        #            mu nu                     b; mu in a; nu
        #
        #                  b*
        #         ----- d P
        #  b       \       i mu   b   b
        # A     =   )   ------- dH   P
        #  mu nu   /    d R       ij  j nu
        #         -----    b mu
        #           ij
        #
        self.timer.start('Atomic Hamiltonian force')
        Fatom_av = np.zeros_like(Fatom_av)
        for u, kpt in enumerate(self.kpt_u):
            for b in self.my_atom_indices:
                H_ii = np.asarray(unpack(self.dH_asp[b][kpt.s]), self.dtype)
                if len(H_ii) == 0:
                    # gemmdot does not like empty matrices!
                    # (has been fixed in the new code)
                    continue
                HP_iM = gemmdot(H_ii, np.ascontiguousarray(
                                self.P_aqMi[b][kpt.q].T.conj()))
                for v in range(3):
                    dPdR_Mi = \
                        self.dPdR_aqvMi[b][kpt.q][v][self.Mstart:self.Mstop]
                    ArhoT_MM = \
                        (gemmdot(dPdR_Mi, HP_iM) * self.rhoT_uMM[u]).real
                    for a, M1, M2 in self.slices():
                        dE = 2 * ArhoT_MM[M1:M2].sum()
                        Fatom_av[a, v] += dE  # the "b; mu in a; nu" term
                        Fatom_av[b, v] -= dE  # the "mu nu" term
        self.timer.stop('Atomic Hamiltonian force')

        return Fatom_av

    def get_den_mat_block_blacs(self, f_n, C_nM, redistributor):
        rho1_mm = self.ksl.calculate_blocked_density_matrix(f_n,
                                                            C_nM).conj()
        rho_mm = redistributor.redistribute(rho1_mm)
        return rho_mm

    def get_pot_term_blacs(self):
        Fpot_av = np.zeros_like(self.Fref_av)
        from gpaw.blacs import BlacsGrid, Redistributor
        self.grid = BlacsGrid(self.ksl.block_comm, self.gd.comm.size,
                              self.bd.comm.size)
        self.blocksize1 = -(-self.nao // self.grid.nprow)
        self.blocksize2 = -(-self.nao // self.grid.npcol)
        desc = self.grid.new_descriptor(self.nao, self.nao,
                                        self.blocksize1, self.blocksize2)
        vt_sG = self.hamiltonian.vt_sG
        self.rhoT_umm = []
        self.ET_umm = []
        self.redistributor = Redistributor(self.grid.comm,
                                           self.ksl.mmdescriptor, desc)
        Fpot_av = np.zeros_like(self.Fref_av)
        for u, kpt in enumerate(self.kpt_u):
            self.timer.start('Get density matrix')
            rhoT_mm = self.get_den_mat_block_blacs(kpt.f_n, kpt.C_nM,
                                                   self.redistributor)
            self.rhoT_umm.append(rhoT_mm)
            self.timer.stop('Get density matrix')
            self.timer.start('Potential')
            rhoT_mM = self.ksl.distribute_to_columns(rhoT_mm, desc)
            vt_G = vt_sG[kpt.s]
            Fpot_av += self.bfs.calculate_force_contribution(vt_G, rhoT_mM,
                                                             kpt.q)
            del rhoT_mM
            self.timer.stop('Potential')

        return Fpot_av

    def get_kin_and_den_term_blacs(self):
        Fkin_av_sum = np.zeros_like(self.Fref_av)
        Ftheta_av_sum = np.zeros_like(self.Fref_av)
        # pcutoff_a = [max([pt.get_cutoff() for pt in setup.pt_j])
        #              for setup in self.setups]
        # phicutoff_a = [max([phit.get_cutoff() for phit in setup.phit_j])
        #                for setup in self.setups]
        # XXX should probably use bdsize x gdsize instead
        # That would be consistent with some existing grids
        # I'm not sure if this is correct
        # XXX what are rows and columns actually?
        dH_asp = self.hamiltonian.dH_asp
        self.timer.start('Get density matrix')
        for kpt in self.kpt_u:
            ET_mm = self.get_den_mat_block_blacs(kpt.f_n * kpt.eps_n, kpt.C_nM,
                                                 self.redistributor)
            self.ET_umm.append(ET_mm)
        self.timer.stop('Get density matrix')
        self.M1start = self.blocksize1 * self.grid.myrow
        self.M2start = self.blocksize2 * self.grid.mycol
        self.M1stop = min(self.M1start + self.blocksize1, self.nao)
        self.M2stop = min(self.M2start + self.blocksize2, self.nao)
        self.m1max = self.M1stop - self.M1start
        self.m2max = self.M2stop - self.M2start
        # from gpaw.lcao.overlap import TwoCenterIntegralCalculator
        self.timer.start('Prepare TCI loop')
        self.M_a = self.bfs.M_a
        Fkin2_av = np.zeros_like(self.Fref_av)
        Ftheta2_av = np.zeros_like(self.Fref_av)
        self.atompairs = self.newtci.a1a2.get_atompairs()
        self.timer.start('broadcast dH')
        self.alldH_asp = {}
        for a in range(len(self.setups)):
            gdrank = self.bfs.sphere_a[a].rank
            if gdrank == self.gd.rank:
                dH_sp = dH_asp[a]
            else:
                ni = self.setups[a].ni
                dH_sp = np.empty((self.nspins, ni * (ni + 1) // 2))
            self.gd.comm.broadcast(dH_sp, gdrank)
            # okay, now everyone gets copies of dH_sp
            self.alldH_asp[a] = dH_sp
        self.timer.stop('broadcast dH')
        # This will get sort of hairy.  We need to account for some
        # three-center overlaps, such as:
        #
        #         a1
        #      Phi   ~a3    a3  ~a3     a2     a2,a1
        #   < ----  |p  > dH   <p   |Phi  > rho
        #      dR
        #
        # To this end we will loop over all pairs of atoms (a1, a3),
        # and then a sub-loop over (a3, a2).
        self.timer.stop('Prepare TCI loop')
        self.timer.start('Not so complicated loop')
        for (a1, a2) in self.atompairs:
            if a1 >= a2:
                # Actually this leads to bad load balance.
                # We should take a1 > a2 or a1 < a2 equally many times.
                # Maybe decide which of these choices
                # depending on whether a2 % 1 == 0
                continue
            m1start = self.M_a[a1] - self.M1start
            m2start = self.M_a[a2] - self.M2start
            if m1start >= self.blocksize1 or m2start >= self.blocksize2:
                continue  # (we have only one block per CPU)
            nm1 = self.setups[a1].nao
            nm2 = self.setups[a2].nao
            m1stop = min(m1start + nm1, self.m1max)
            m2stop = min(m2start + nm2, self.m2max)
            if m1stop <= 0 or m2stop <= 0:
                continue
            m1start = max(m1start, 0)
            m2start = max(m2start, 0)
            J1start = max(0, self.M1start - self.M_a[a1])
            J2start = max(0, self.M2start - self.M_a[a2])
            M1stop = J1start + m1stop - m1start
            J2stop = J2start + m2stop - m2start
            dThetadR_qvmm, dTdR_qvmm = self.newtci.dOdR_dTdR(a1, a2)
            for u, kpt in enumerate(self.kpt_u):
                rhoT_mm = self.rhoT_umm[u][m1start:m1stop, m2start:m2stop]
                ET_mm = self.ET_umm[u][m1start:m1stop, m2start:m2stop]
                Fkin_v = 2.0 * (dTdR_qvmm[kpt.q][:, J1start:M1stop,
                                                 J2start:J2stop] *
                                rhoT_mm[np.newaxis]).real.sum(-1).sum(-1)
                Ftheta_v = 2.0 * (dThetadR_qvmm[kpt.q][:, J1start:M1stop,
                                                       J2start:J2stop] *
                                  ET_mm[np.newaxis]).real.sum(-1).sum(-1)
                Fkin2_av[a1] += Fkin_v
                Fkin2_av[a2] -= Fkin_v
                Ftheta2_av[a1] -= Ftheta_v
                Ftheta2_av[a2] += Ftheta_v
        Fkin_av = Fkin2_av
        Ftheta_av = Ftheta2_av
        self.timer.stop('Not so complicated loop')

        Fkin_av_sum += Fkin_av
        Ftheta_av_sum += Ftheta_av

        return Fkin_av_sum, Ftheta_av_sum

    def get_at_den_and_den_paw_blacs(self):
        Fatom_av = np.zeros_like(self.Fref_av)
        Frho_av = np.zeros_like(self.Fref_av)
        Fatom_av_sum = np.zeros_like(self.Fref_av)
        Frho_av_sum = np.zeros_like(self.Fref_av)
        self.dHP_and_dSP_aauim = {}
        self.a2values = {}
        for (a2, a3) in self.atompairs:
            if a3 not in self.a2values:
                self.a2values[a3] = []
            self.a2values[a3].append(a2)

        self.timer.start('Complicated loop')
        for a1, a3 in self.atompairs:
            if a1 == a3:
                # Functions reside on same atom, so their overlap
                # does not change when atom is displaced
                continue
            m1start = self.M_a[a1] - self.M1start
            if m1start >= self.blocksize1:
                continue
            nm1 = self.setups[a1].nao
            m1stop = min(m1start + nm1, self.m1max)
            if m1stop <= 0:
                continue
            dPdR_qvim = self.newtci.dPdR(a3, a1)
            if dPdR_qvim is None:
                continue
            dPdR_qvmi = -dPdR_qvim.transpose(0, 1, 3, 2).conj()
            m1start = max(m1start, 0)
            J1start = max(0, self.M1start - self.M_a[a1])
            J1stop = J1start + m1stop - m1start
            dPdR_qvmi = dPdR_qvmi[:, :, J1start:J1stop, :].copy()
            for a2 in self.a2values[a3]:
                m2start = self.M_a[a2] - self.M2start
                if m2start >= self.blocksize2:
                    continue
                nm2 = self.setups[a2].nao
                m2stop = min(m2start + nm2, self.m2max)
                if m2stop <= 0:
                    continue
                m2start = max(m2start, 0)
                J2start = max(0, self.M2start - self.M_a[a2])
                J2stop = J2start + m2stop - m2start
                if (a2, a3) in self.dHP_and_dSP_aauim:
                    dHP_uim, dSP_uim = self.dHP_and_dSP_aauim[(a2, a3)]
                else:
                    P_qim = self.newtci.P(a3, a2)
                    if P_qim is None:
                        continue
                    P_qmi = P_qim.transpose(0, 2, 1).conj()
                    P_qmi = P_qmi[:, J2start:J2stop].copy()
                    dH_sp = self.alldH_asp[a3]
                    dS_ii = self.setups[a3].dO_ii
                    dHP_uim = []
                    dSP_uim = []
                    for u, kpt in enumerate(self.kpt_u):
                        dH_ii = unpack(dH_sp[kpt.s])
                        dHP_im = np.dot(P_qmi[kpt.q], dH_ii).T.conj()
                        # XXX only need nq of these,
                        # but the looping is over all u
                        dSP_im = np.dot(P_qmi[kpt.q], dS_ii).T.conj()
                        dHP_uim.append(dHP_im)
                        dSP_uim.append(dSP_im)
                        self.dHP_and_dSP_aauim[(a2, a3)] = dHP_uim, dSP_uim
                for u, kpt in enumerate(self.kpt_u):
                    rhoT_mm = self.rhoT_umm[u][m1start:m1stop, m2start:m2stop]
                    ET_mm = self.ET_umm[u][m1start:m1stop, m2start:m2stop]
                    dPdRdHP_vmm = np.dot(dPdR_qvmi[kpt.q], dHP_uim[u])
                    dPdRdSP_vmm = np.dot(dPdR_qvmi[kpt.q], dSP_uim[u])
                    Fatom_c = 2.0 * (dPdRdHP_vmm *
                                     rhoT_mm).real.sum(-1).sum(-1)
                    Frho_c = 2.0 * (dPdRdSP_vmm *
                                    ET_mm).real.sum(-1).sum(-1)
                    Fatom_av[a1] += Fatom_c
                    Fatom_av[a3] -= Fatom_c
                    Frho_av[a1] -= Frho_c
                    Frho_av[a3] += Frho_c
        self.timer.stop('Complicated loop')

        Fatom_av_sum += Fatom_av
        Frho_av_sum += Frho_av

        return Fatom_av_sum, Frho_av_sum



from math import pi

import numpy as np

from gpaw.xc.lda import calculate_paw_correction
from gpaw.utilities.blas import axpy
from gpaw.fd_operators import Gradient
from gpaw.sphere.lebedev import Y_nL, weight_n
from gpaw.xc.pawcorrection import rnablaY_nLv
from gpaw.xc.functional import XCFunctional


class GGARadialExpansion:
    def __init__(self, rcalc, *args):
        self.rcalc = rcalc
        self.args = args

    def __call__(self, rgd, D_sLq, n_qg, nc0_sg):
        n_sLg = np.dot(D_sLq, n_qg)
        n_sLg[:, 0] += nc0_sg

        dndr_sLg = np.empty_like(n_sLg)
        for n_Lg, dndr_Lg in zip(n_sLg, dndr_sLg):
            for n_g, dndr_g in zip(n_Lg, dndr_Lg):
                rgd.derivative(n_g, dndr_g)

        nspins, Lmax, nq = D_sLq.shape
        dEdD_sqL = np.zeros((nspins, nq, Lmax))

        E = 0.0
        for n, Y_L in enumerate(Y_nL[:, :Lmax]):
            w = weight_n[n]
            rnablaY_Lv = rnablaY_nLv[n, :Lmax]
            e_g, dedn_sg, b_vsg, dedsigma_xg = \
                self.rcalc(rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv, n,
                           *self.args)
            dEdD_sqL += np.dot(rgd.dv_g * dedn_sg,
                               n_qg.T)[:, :, np.newaxis] * (w * Y_L)
            dedsigma_xg *= rgd.dr_g
            B_vsg = dedsigma_xg[::2] * b_vsg
            if nspins == 2:
                B_vsg += 0.5 * dedsigma_xg[1] * b_vsg[:, ::-1]
            B_vsq = np.dot(B_vsg, n_qg.T)
            dEdD_sqL += 8 * pi * w * np.inner(rnablaY_Lv, B_vsq.T).T
            E += w * rgd.integrate(e_g)

        return E, dEdD_sqL


# First part of gga_calculate_radial - initializes some quantities.
def radial_gga_vars(rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv):
    nspins = len(n_sLg)

    n_sg = np.dot(Y_L, n_sLg)

    a_sg = np.dot(Y_L, dndr_sLg)
    b_vsg = np.dot(rnablaY_Lv.T, n_sLg)

    sigma_xg = rgd.empty(2 * nspins - 1)
    sigma_xg[::2] = (b_vsg ** 2).sum(0)
    if nspins == 2:
        sigma_xg[1] = (b_vsg[:, 0] * b_vsg[:, 1]).sum(0)
    sigma_xg[:, 1:] /= rgd.r_g[1:] ** 2
    sigma_xg[:, 0] = sigma_xg[:, 1]
    sigma_xg[::2] += a_sg ** 2
    if nspins == 2:
        sigma_xg[1] += a_sg[0] * a_sg[1]

    e_g = rgd.empty()
    dedn_sg = rgd.zeros(nspins)
    dedsigma_xg = rgd.zeros(2 * nspins - 1)
    return e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg, b_vsg


def add_radial_gradient_correction(rgd, sigma_xg, dedsigma_xg, a_sg):
    nspins = len(a_sg)
    vv_sg = sigma_xg[:nspins]  # reuse array
    for s in range(nspins):
        rgd.derivative2(-2 * rgd.dv_g * dedsigma_xg[2 * s] * a_sg[s],
                        vv_sg[s])
    if nspins == 2:
        v_g = sigma_xg[2]
        rgd.derivative2(rgd.dv_g * dedsigma_xg[1] * a_sg[1], v_g)
        vv_sg[0] -= v_g
        rgd.derivative2(rgd.dv_g * dedsigma_xg[1] * a_sg[0], v_g)
        vv_sg[1] -= v_g

    vv_sg[:, 1:] /= rgd.dv_g[1:]
    vv_sg[:, 0] = vv_sg[:, 1]
    return vv_sg


class GGARadialCalculator:
    def __init__(self, kernel):
        self.kernel = kernel

    def __call__(self, rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv, n):
        (e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg,
         b_vsg) = radial_gga_vars(rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv)
        self.kernel.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        vv_sg = add_radial_gradient_correction(rgd, sigma_xg,
                                               dedsigma_xg, a_sg)
        return e_g, dedn_sg + vv_sg, b_vsg, dedsigma_xg


def calculate_sigma(gd, grad_v, n_sg):
    r"""Calculate sigma(r) and grad n(r).
                  _     __   _  2     __    _
    Returns sigma(r) = |\/ n(r)|  and \/ n (r).

    With multiple spins, sigma has the three elements

                   _     __     _  2
            sigma (r) = |\/ n  (r)|  ,
                 0           up

                   _     __     _      __     _
            sigma (r) =  \/ n  (r)  .  \/ n  (r) ,
                 1           up            dn

                   _     __     _  2
            sigma (r) = |\/ n  (r)|  .
                 2           dn
    """
    nspins = len(n_sg)
    gradn_svg = gd.empty((nspins, 3))
    sigma_xg = gd.zeros(nspins * 2 - 1)
    for v in range(3):
        for s in range(nspins):
            grad_v[v](n_sg[s], gradn_svg[s, v])
            axpy(1.0, gradn_svg[s, v]**2, sigma_xg[2 * s])
        if nspins == 2:
            axpy(1.0, gradn_svg[0, v] * gradn_svg[1, v], sigma_xg[1])
    return sigma_xg, gradn_svg


def add_gradient_correction(grad_v, gradn_svg, sigma_xg, dedsigma_xg, v_sg):
    r"""Add gradient correction to potential.

    ::

                      __   /    de(r)    __      \
        v  (r) += -2  \/ . |  ---------  \/ n(r) |
         xc                \  dsigma(r)          /

    Adds arbitrary data to sigma_xg.  Be sure to pass a copy if you need
    sigma_xg after calling this function.
    """
    nspins = len(v_sg)
    # vv_g is a calculation buffer.  Its contents will be corrupted.
    vv_g = sigma_xg[0]
    for v in range(3):
        for s in range(nspins):
            grad_v[v](dedsigma_xg[2 * s] * gradn_svg[s, v], vv_g)
            axpy(-2.0, vv_g, v_sg[s])
            if nspins == 2:
                grad_v[v](dedsigma_xg[1] * gradn_svg[s, v], vv_g)
                axpy(-1.0, vv_g, v_sg[1 - s])
                # TODO: can the number of gradient evaluations be reduced?


def gga_vars(gd, grad_v, n_sg):
    nspins = len(n_sg)
    sigma_xg, gradn_svg = calculate_sigma(gd, grad_v, n_sg)
    dedsigma_xg = gd.empty(nspins * 2 - 1)
    return sigma_xg, dedsigma_xg, gradn_svg


def get_gradient_ops(gd, nn):
    return [Gradient(gd, v, n=nn).apply for v in range(3)]


class GGA(XCFunctional):
    def __init__(self, kernel, stencil=2):
        XCFunctional.__init__(self, kernel.name, kernel.type)
        self.kernel = kernel
        self.stencil_range = stencil

    def set_grid_descriptor(self, gd):
        XCFunctional.set_grid_descriptor(self, gd)
        self.grad_v = get_gradient_ops(gd, self.stencil_range)

    def todict(self):
        d = super(GGA, self).todict()
        d['stencil'] = self.stencil_range
        return d

    def get_description(self):
        return ('{} with {} nearest neighbor stencil'
                .format(self.name, self.stencil_range))

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        sigma_xg, dedsigma_xg, gradn_svg = gga_vars(gd, self.grad_v, n_sg)
        self.kernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        add_gradient_correction(self.grad_v, gradn_svg, sigma_xg,
                                dedsigma_xg, v_sg)

    def calculate_paw_correction(self, setup, D_sp, dEdD_sp=None,
                                 addcoredensity=True, a=None):
        rcalc = GGARadialCalculator(self.kernel)
        expansion = GGARadialExpansion(rcalc)
        return calculate_paw_correction(expansion,
                                        setup, D_sp, dEdD_sp,
                                        addcoredensity, a)

    def stress_tensor_contribution(self, n_sg):
        sigma_xg, gradn_svg = calculate_sigma(self.gd, self.grad_v, n_sg)
        nspins = len(n_sg)
        dedsigma_xg = self.gd.empty(nspins * 2 - 1)
        v_sg = self.gd.zeros(nspins)
        e_g = self.gd.empty()
        self.kernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)

        def integrate(a1_g, a2_g=None):
            return self.gd.integrate(a1_g, a2_g, global_integral=False)

        P = integrate(e_g)
        for v_g, n_g in zip(v_sg, n_sg):
            P -= integrate(v_g, n_g)
        for sigma_g, dedsigma_g in zip(sigma_xg, dedsigma_xg):
            P -= 2 * integrate(sigma_g, dedsigma_g)

        stress_vv = P * np.eye(3)
        for v1 in range(3):
            for v2 in range(3):
                stress_vv[v1, v2] -= integrate(gradn_svg[0, v1] *
                                               gradn_svg[0, v2],
                                               dedsigma_xg[0]) * 2
                if nspins == 2:
                    stress_vv[v1, v2] -= integrate(gradn_svg[0, v1] *
                                                   gradn_svg[1, v2],
                                                   dedsigma_xg[1]) * 2
                    stress_vv[v1, v2] -= integrate(gradn_svg[1, v1] *
                                                   gradn_svg[1, v2],
                                                   dedsigma_xg[2]) * 2
        self.gd.comm.sum(stress_vv)
        return stress_vv

    def calculate_spherical(self, rgd, n_sg, v_sg, e_g=None):
        dndr_sg = np.empty_like(n_sg)
        for n_g, dndr_g in zip(n_sg, dndr_sg):
            rgd.derivative(n_g, dndr_g)
        if e_g is None:
            e_g = rgd.empty()

        rcalc = GGARadialCalculator(self.kernel)

        e_g[:], dedn_sg = rcalc(rgd, n_sg[:, np.newaxis],
                                [1.0],
                                dndr_sg[:, np.newaxis],
                                np.zeros((1, 3)), n=None)[:2]
        v_sg[:] = dedn_sg
        return rgd.integrate(e_g)


class PurePythonGGAKernel:
    def __init__(self, name):
        self.type = 'GGA'
        self.name, self.kappa, self.mu, self.beta = pbe_constants(name)

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg, dedsigma_xg,
                  tau_sg=None, dedtau_sg=None):
        e_g[:] = 0.
        dedsigma_xg[:] = 0.

        # spin-paired:
        if len(n_sg) == 1:
            n = n_sg[0]
            n[n < 1e-20] = 1e-40

            # exchange
            res = gga_x(self.name, 0, n, sigma_xg[0], self.kappa, self.mu)
            ex, rs, dexdrs, dexda2 = res
            # correlation
            res = gga_c(self.name, 0, n, sigma_xg[0], 0, self.beta)
            ec, rs_, decdrs, decda2, decdzeta = res

            e_g[:] += n * (ex + ec)
            dedn_sg[:] += ex + ec - rs * (dexdrs + decdrs) / 3.
            dedsigma_xg[:] += n * (dexda2 + decda2)

        # spin-polarized:
        else:
            na = 2. * n_sg[0]
            na[na < 1e-20] = 1e-40

            nb = 2. * n_sg[1]
            nb[nb < 1e-20] = 1e-40

            n = 0.5 * (na + nb)
            zeta = 0.5 * (na - nb) / n

            # exchange
            exa, rsa, dexadrs, dexada2 = gga_x(
                self.name, 1, na, 4.0 * sigma_xg[0], self.kappa, self.mu)
            exb, rsb, dexbdrs, dexbda2 = gga_x(
                self.name, 1, nb, 4.0 * sigma_xg[2], self.kappa, self.mu)
            a2 = sigma_xg[0] + 2.0 * sigma_xg[1] + sigma_xg[2]
            # correlation
            ec, rs, decdrs, decda2, decdzeta = gga_c(
                self.name, 1, n, a2, zeta, self.beta)

            e_g[:] += 0.5 * (na * exa + nb * exb) + n * ec
            dedn_sg[0][:] += (exa + ec - (rsa * dexadrs + rs * decdrs) / 3.0
                              - (zeta - 1.0) * decdzeta)
            dedn_sg[1][:] += (exb + ec - (rsb * dexbdrs + rs * decdrs) / 3.0
                              - (zeta + 1.0) * decdzeta)
            dedsigma_xg[0][:] += 2.0 * na * dexada2 + n * decda2
            dedsigma_xg[1][:] += 2.0 * n * decda2
            dedsigma_xg[2][:] += 2.0 * nb * dexbda2 + n * decda2


def pbe_constants(name):
    if name == 'pyPBE':
        name = 'PBE'
        kappa = 0.804
        mu = 0.2195149727645171
        beta = 0.06672455060314922
    elif name in ['pyPBEsol', 'pyzvPBEsol']:
        name = name[2:]
        kappa = 0.804
        mu = 10. / 81.
        beta = 0.046
    elif name == 'pyRPBE':
        name = 'RPBE'
        kappa = 0.804
        mu = 0.2195149727645171
        beta = 0.06672455060314922
    else:
        raise NotImplementedError(name)

    return name, kappa, mu, beta


# a2 = |grad n|^2
def gga_x(name, spin, n, a2, kappa, mu, dedmu_g=None):
    # if dedmu is given, calculate also d(e_x)/d(mu)
    assert spin in [0, 1]

    C0I, C1, C2, C3, CC1, CC2, IF2, GAMMA = gga_constants()
    rs = (C0I / n)**(1 / 3.)

    # lda part
    ex = C1 / rs
    dexdrs = -ex / rs

    # gga part
    c = (C2 * rs / n)**2.
    s2 = a2 * c

    if name in ['PBE', 'PBEsol', 'zvPBEsol', 'QNA']:
        x = 1.0 + mu * s2 / kappa
        Fx = 1.0 + kappa - kappa / x
        dFxds2 = mu / (x**2.)
    elif name == 'RPBE':
        arg = np.maximum(-mu * s2 / kappa, -5.e2)
        x = np.exp(arg)
        Fx = 1.0 + kappa * (1.0 - x)
        dFxds2 = mu * x
    else:
        raise NotImplementedError

    ds2drs = 8.0 * c * a2 / rs
    dexdrs = dexdrs * Fx + ex * dFxds2 * ds2drs
    dexda2 = ex * dFxds2 * c
    if dedmu_g is not None:
        dedmu_g[:] = ex * s2 / (1 + mu * s2 / kappa)**2
    ex *= Fx

    return ex, rs, dexdrs, dexda2


def gga_c(name, spin, n, a2, zeta, BETA, decdbeta_g=None):
    # If decdbeta is specified, calculate also d(e_c)/d(beta)
    assert spin in [0, 1]
    from gpaw.xc.lda import G

    C0I, C1, C2, C3, CC1, CC2, IF2, GAMMA = gga_constants()
    rs = (C0I / n)**(1 / 3.)

    if name == 'zvPBEsol':
        zv_a = 1.8
        zv_o = 9. / 2.
        zv_x = 1. / 6.

    # lda part
    ec, decdrs_0 = G(rs**0.5, 0.031091, 0.21370, 7.5957,
                     3.5876, 1.6382, 0.49294)

    if spin == 0:
        decdrs = decdrs_0
        decdzeta = 0.  # dummy
    else:
        e1, decdrs_1 = G(rs**0.5, 0.015545, 0.20548, 14.1189,
                         6.1977, 3.3662, 0.62517)
        alpha, dalphadrs = G(rs**0.5, 0.016887, 0.11125, 10.357,
                             3.6231, 0.88026, 0.49671)
        alpha *= -1.
        dalphadrs *= -1.
        zp = 1.0 + zeta
        zm = 1.0 - zeta
        xp = zp**(1 / 3.)
        xm = zm**(1 / 3.)
        f = CC1 * (zp * xp + zm * xm - 2.0)
        f1 = CC2 * (xp - xm)
        zeta3 = zeta * zeta * zeta
        zeta4 = zeta * zeta * zeta * zeta
        x = 1.0 - zeta4
        decdrs = (decdrs_0 * (1.0 - f * zeta4) +
                  decdrs_1 * f * zeta4 +
                  dalphadrs * f * x * IF2)
        decdzeta = (4.0 * zeta3 * f * (e1 - ec - alpha * IF2) +
                    f1 * (zeta4 * e1 - zeta4 * ec + x * alpha * IF2))
        ec += alpha * IF2 * f * x + (e1 - ec) * f * zeta4

    # gga part
    n2 = n * n
    if spin == 1:
        phi = 0.5 * (xp * xp + xm * xm)
        phi2 = phi * phi
        phi3 = phi * phi2
        t2 = C3 * a2 * rs / (n2 * phi2)
        y = -ec / (GAMMA * phi3)
        if name == 'zvPBEsol':
            u3 = t2**(3. / 2.) * phi3 * (rs / 3.)**(-3. * zv_x)
            zvarg = -zv_a * u3 * abs(zeta)**zv_o
            zvf = np.exp(zvarg)
    else:
        t2 = C3 * a2 * rs / n2
        y = -ec / GAMMA

    x = np.exp(y)

    A = np.zeros_like(x)
    indices = np.nonzero(y)
    if isinstance(BETA, np.ndarray):
        A[indices] = (BETA[indices] / (GAMMA * (x[indices] - 1.0)))
    else:
        A[indices] = (BETA / (GAMMA * (x[indices] - 1.0)))

    At2 = A * t2
    nom = 1.0 + At2
    denom = nom + At2 * At2
    X = 1.0 + BETA * t2 * nom / (denom * GAMMA)
    H = GAMMA * np.log(X)
    tmp = (GAMMA * BETA / (denom * (BETA * t2 * nom + GAMMA * denom)))
    tmp2 = A * A * x / BETA
    dAdrs = tmp2 * decdrs
    if spin == 1:
        H *= phi3
        tmp *= phi3
        dAdrs /= phi3
        if name == 'zvPBEsol':
            H_ = H.copy()
            H *= zvf
    dHdt2 = (1.0 + 2.0 * At2) * tmp
    dHdA = -At2 * t2 * t2 * (2.0 + At2) * tmp
    decdrs += dHdt2 * 7.0 * t2 / rs + dHdA * dAdrs
    decda2 = dHdt2 * C3 * rs / n2
    if spin == 1:
        dphidzeta = np.zeros_like(x)
        ind1 = np.nonzero(xp)
        ind2 = np.nonzero(xm)
        dphidzeta[ind1] += 1.0 / (3.0 * xp[ind1])
        dphidzeta[ind2] -= 1.0 / (3.0 * xm[ind2])
        dAdzeta = tmp2 * (decdzeta - 3.0 * ec * dphidzeta / phi) / phi3
        decdzeta += ((3.0 * H / phi - dHdt2 * 2.0 * t2 / phi) * dphidzeta
                     + dHdA * dAdzeta)
        decda2 /= phi2
        if name == 'zvPBEsol':
            u3_ = (t2**(3. / 2.) *
                   phi3 * rs**(-3. * zv_x - 1.) / 3.**(-3. * zv_x))
            zvarg_ = -zv_a * u3_ * abs(zeta)**zv_o
            dzvfdrs = -3. * zv_x * zvf * zvarg_
            decdrs *= zvf
            decdrs += H_ * dzvfdrs

            dt2da2 = C3 * rs / n2
            assert np.shape(dt2da2) == np.shape(t2)
            dzvfda2 = dt2da2 * zvf * zvarg * 3. / (2. * t2)
            decda2 *= zvf
            decda2 += H_ * dzvfda2

            dadz = zv_o * abs(zeta)**(zv_o - 1.)
            dbdphi = 3. * u3 / phi
            dPdzeta = u3 * dadz + abs(zeta)**(zv_o) * dbdphi * dphidzeta
            dzvfdzeta = -zv_a * zvf * dPdzeta
            decdzeta *= zvf
            decdzeta += H_ * dzvfdzeta
    ec += H

    if decdbeta_g is not None:
        if spin == 0:
            phi3 = 1.0
        Y = GAMMA * phi3
        decdbeta_g[:] = Y / X * t2 / GAMMA
        decdbeta_g *= ((1 + 2 * At2) / (1 + At2 + At2**2) -
                       (1 + At2) * (At2 + 2 * At2**2) / (1 + At2 + At2**2)**2)

    return ec, rs, decdrs, decda2, decdzeta


def gga_constants():
    from gpaw.xc.lda import lda_constants
    C0I, C1, CC1, CC2, IF2 = lda_constants()
    C2 = 0.26053088059892404
    C3 = 0.10231023756535741
    GAMMA = 0.0310906908697

    return C0I, C1, C2, C3, CC1, CC2, IF2, GAMMA



import os


class CLICommand:
    """Manipulate/show content of GPAW-restart file."""

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('gpw', metavar='gpw-file')
        parser.add_argument('-w', '--remove-wave-functions',
                            action='store_true')

    @staticmethod
    def run(args):
        if args.remove_wave_functions:
            import ase.io.ulm as ulm
            reader = ulm.open(args.gpw)
            if 'values' not in reader.wave_functions:
                print('No wave functions in', args.gpw)
            else:
                ulm.copy(reader, args.gpw + '.temp',
                         exclude={'.wave_functions.values'})
                reader.close()
                os.rename(args.gpw + '.temp', args.gpw)
        else:
            from gpaw import GPAW
            GPAW(args.gpw)



from gpaw.gpu import cupy as cp
import numpy as np


class CuPyMPI:
    """Quick'n'dirty wrapper to make things work without a GPU-aware MPI."""
    def __init__(self, comm):
        self.comm = comm
        self.rank = comm.rank
        self.size = comm.size

    def sum(self, array, root=-1):
        if isinstance(array, (float, int)):
            return self.comm.sum(array, root)
        if isinstance(array, np.ndarray):
            self.comm.sum(array, root)
            return
        a = array.get()
        self.comm.sum(a, root)
        array.set(a)

    def sum_scalar(self, a, root=-1):
        return self.comm.sum_scalar(a, root)

    def min_scalar(self, a, root=-1):
        return self.comm.min_scalar(a, root)

    def max(self, array):
        self.comm.max(array)

    def all_gather(self, a, b):
        self.comm.all_gather(a, b)

    def gather(self, a, rank, b):
        if isinstance(a, np.ndarray):
            self.comm.gather(a, rank, b)
        else:
            if rank == self.rank:
                c = np.empty(b.shape, b.dtype)
            else:
                c = None
            self.comm.gather(a.get(), rank, c)
            if rank == self.rank:
                b[:] = cp.asarray(c)

    def scatter(self, fro, to, root=0):
        if isinstance(to, np.ndarray):
            self.comm.scatter(fro, to, root)
            return
        b = np.empty(to.shape, to.dtype)
        if self.rank == root:
            a = fro.get()
        else:
            a = None
        self.comm.scatter(a, b, root)
        to[:] = cp.asarray(b)

    def broadcast(self, a, root):
        if isinstance(a, np.ndarray):
            self.comm.broadcast(a, root)
            return
        b = a.get()
        self.comm.broadcast(b, root)
        a[:] = cp.asarray(b)

    def receive(self, a, rank, tag=0):
        if isinstance(a, np.ndarray):
            return self.comm.receive(a, rank, tag)
        b = np.empty(a.shape, a.dtype)
        self.comm.receive(b, rank, tag)
        a[:] = cp.asarray(b)

    def ssend(self, a, rank, tag):
        if isinstance(a, np.ndarray):
            self.comm.ssend(a, rank, tag)
        else:
            self.comm.ssend(a.get(), rank, tag)

    def send(self, a, rank, tag=0, block=True):
        if isinstance(a, np.ndarray):
            return self.comm.send(a, rank, tag, block)
        b = a.get()
        request = self.comm.send(b, rank, tag, block)
        if not block:
            return CuPyRequest(request, b)

    def alltoallv(self,
                  fro, ssizes, soffsets,
                  to, rsizes, roffsets):
        a = np.empty(to.shape, to.dtype)
        self.comm.alltoallv(fro.get(), ssizes, soffsets,
                            a, rsizes, roffsets)
        to[:] = cp.asarray(a)

    def wait(self, request):
        self.comm.wait(request.request)

    def get_c_object(self):
        return self.comm.get_c_object()


class CuPyRequest:
    def __init__(self, request, array):
        self.request = request
        self.array = array



import numpy as np

from gpaw.blacs import BlacsGrid
from gpaw.blacs import Redistributor


def collect_uX(kd, comm, a_uX, s, k):
    # Comm is a communicator orthogonal to kd.comm (ie, domainband_comm)
    Xshape = a_uX[0].shape
    dtype = a_uX[0].dtype
    kpt_rank, q = kd.get_rank_and_index(k)
    u = q * kd.nspins + s
    if kd.comm.rank == kpt_rank:
        a_X = a_uX[u]
        # Comm master send to the global master
        if comm.rank == 0:
            if kpt_rank == 0:
                # assert world.rank == 0
                return a_X
            else:
                kd.comm.ssend(a_X, 0, 2018)
    elif comm.rank == 0 and kpt_rank != 0:
        # assert world.rank == 0
        a_X = np.empty(Xshape, dtype=dtype)
        kd.comm.receive(a_X, kpt_rank, 2018)
        return a_X


def write_uX(kd, comm, writer, name, a_uX):
    ushape = (kd.nspins, kd.nibzkpts)
    Xshape = a_uX[0].shape
    dtype = a_uX[0].dtype
    writer.add_array(name, ushape + Xshape, dtype=dtype)
    for s in range(kd.nspins):
        for k in range(kd.nibzkpts):
            a_X = collect_uX(kd, comm, a_uX, s, k)
            writer.fill(a_X)


def read_uX(kpt_u, reader, name):
    a_uX = []
    for kpt in kpt_u:
        indices = (kpt.s, kpt.k)
        # TODO: does this read on all the comm ranks in vain?
        a_X = reader.proxy(name, *indices)[:]
        a_uX.append(a_X)
    return a_uX


def distribute_nM(ksl, a_nM):
    if not ksl.using_blacs:
        return a_nM

    dtype = a_nM.dtype
    ksl.nMdescriptor.checkassert(a_nM)
    if ksl.gd.rank != 0:
        a_nM = ksl.nM_unique_descriptor.zeros(dtype=dtype)

    nM2mm = Redistributor(ksl.block_comm, ksl.nM_unique_descriptor,
                          ksl.mmdescriptor)

    a_mm = ksl.mmdescriptor.empty(dtype=dtype)
    nM2mm.redistribute(a_nM, a_mm, ksl.bd.nbands, ksl.nao)
    return a_mm


def collect_MM(ksl, a_mm):
    if not ksl.using_blacs:
        return a_mm

    dtype = a_mm.dtype
    NM = ksl.nao
    grid = BlacsGrid(ksl.block_comm, 1, 1)
    MM_descriptor = grid.new_descriptor(NM, NM, NM, NM)
    mm2MM = Redistributor(ksl.block_comm,
                          ksl.mmdescriptor,
                          MM_descriptor)

    a_MM = MM_descriptor.empty(dtype=dtype)
    mm2MM.redistribute(a_mm, a_MM)
    return a_MM


def collect_uMM(kd, ksl, a_uMM, s, k):
    return collect_wuMM(kd, ksl, [a_uMM], 0, s, k)


def collect_wuMM(kd, ksl, a_wuMM, w, s, k):
    # This function is based on
    # gpaw/wavefunctions/base.py: WaveFunctions.collect_auxiliary()

    dtype = a_wuMM[0][0].dtype
    NM = ksl.nao
    kpt_rank, q = kd.get_rank_and_index(k)
    u = q * kd.nspins + s
    if kd.comm.rank == kpt_rank:
        a_MM = a_wuMM[w][u]

        # Collect within blacs grid
        a_MM = collect_MM(ksl, a_MM)

        # KSL master send a_MM to the global master
        if ksl.block_comm.rank == 0:
            if kpt_rank == 0:
                assert ksl.world.rank == 0
                # I have it already
                return a_MM
            else:
                kd.comm.send(a_MM, 0, 2017)
                return None
    elif ksl.block_comm.rank == 0 and kpt_rank != 0:
        assert ksl.world.rank == 0
        a_MM = np.empty((NM, NM), dtype=dtype)
        kd.comm.receive(a_MM, kpt_rank, 2017)
        return a_MM


def distribute_MM(ksl, a_MM):
    if not ksl.using_blacs:
        return a_MM

    dtype = a_MM.dtype
    NM = ksl.nao
    grid = BlacsGrid(ksl.block_comm, 1, 1)
    MM_descriptor = grid.new_descriptor(NM, NM, NM, NM)
    MM2mm = Redistributor(ksl.block_comm,
                          MM_descriptor,
                          ksl.mmdescriptor)
    if ksl.block_comm.rank != 0:
        a_MM = MM_descriptor.empty(dtype=dtype)

    a_mm = ksl.mmdescriptor.empty(dtype=dtype)
    MM2mm.redistribute(a_MM, a_mm)
    return a_mm


def write_uMM(kd, ksl, writer, name, a_uMM):
    return write_wuMM(kd, ksl, writer, name, [a_uMM], wlist=[0])


def write_wuMM(kd, ksl, writer, name, a_wuMM, wlist):
    NM = ksl.nao
    dtype = a_wuMM[0][0].dtype
    writer.add_array(name,
                     (len(wlist), kd.nspins, kd.nibzkpts, NM, NM),
                     dtype=dtype)
    for w in wlist:
        for s in range(kd.nspins):
            for k in range(kd.nibzkpts):
                a_MM = collect_wuMM(kd, ksl, a_wuMM, w, s, k)
                writer.fill(a_MM)


def read_uMM(kpt_u, ksl, reader, name):
    return read_wuMM(kpt_u, ksl, reader, name, wlist=[0])[0]


def read_wuMM(kpt_u, ksl, reader, name, wlist):
    a_wuMM = []
    for w in wlist:
        a_uMM = []
        for kpt in kpt_u:
            indices = (w, kpt.s, kpt.k)
            # TODO: does this read on all the ksl ranks in vain?
            a_MM = reader.proxy(name, *indices)[:]
            a_MM = distribute_MM(ksl, a_MM)
            a_uMM.append(a_MM)
        a_wuMM.append(a_uMM)
    return a_wuMM



from math import sqrt

import numpy as np
from ase.atoms import Atoms
from ase.units import Bohr, Hartree
from ase.io.cube import read_cube_data, write_cube
from ase.dft.stm import STM

import gpaw.mpi as mpi
from gpaw.grid_descriptor import GridDescriptor


class SimpleStm(STM):

    """Simple STM object to simulate STM pictures.

    The simulation uses either a single pseudo-wavefunction (PWF)
    or the PWFs inside the given bias range."""

    def __init__(self, atoms):

        self.file = None
        self.is_wf = False
        self.bias = None
        self.ldos = None
        self.heights = None

        if isinstance(atoms, str):
            self.read_3D(atoms)
            self.calc = None
            self.atoms = None
        else:
            if isinstance(atoms, Atoms):
                self.atoms = atoms
                self.calc = atoms.calc
            else:
                self.calc = atoms
                self.atoms = self.calc.get_atoms()
            self.calc.converge_wave_functions()

            self.gd = self.calc.wfs.gd
            self.offset_c = [int(not a) for a in self.gd.pbc_c]

            STM.__init__(self, self.atoms)

    def calculate_ldos(self, bias):
        """bias is the n, k, s list/tuple."""
        if self.calc is None:
            return

        self.bias = bias
        self.is_wf = True
        self.ldos = self.gd.zeros()

        if hasattr(bias, '__len__') and len(bias) == 3:
            n, k, s = bias
            # only a single wf requested
            kd = self.calc.wfs.kd
            rank, q = kd.who_has(k)
            if kd.comm.rank == rank:
                u = q * kd.nspins + s
                self.add_wf_to_ldos(n, u, weight=1)

        else:
            # energy bias
            try:
                efermi_s = self.calc.get_fermi_levels()
            except ValueError:
                efermi_s = np.array([self.calc.get_fermi_level()] * 2)

            if isinstance(bias, (int, float)):
                # bias given
                if bias > 0:
                    # positive bias = negative tip
                    # -> probe unoccupied states
                    emin_s = efermi_s
                    emax_s = efermi_s + bias
                    occupied = False
                else:
                    # negative bias = positive tip
                    # -> probe occupied states
                    emin_s = efermi_s + bias
                    emax_s = efermi_s
                    occupied = True
            else:
                # emin and emax given
                emin, emax = bias
                if abs(emin) > abs(emax):
                    occupied = True
                else:
                    occupied = False
                emin_s = np.array([emin + efermi_s] * 2)
                emax_s = np.array([emax + efermi_s] * 2)

            emin_s /= Hartree
            emax_s /= Hartree

            for u in range(len(self.calc.wfs.kpt_u)):
                kpt = self.calc.wfs.kpt_u[u]
                emin = emin_s[kpt.s]
                emax = emax_s[kpt.s]
                for n, eps in enumerate(kpt.eps_n):
                    if eps > emin and eps < emax:
                        if occupied:
                            weight = kpt.f_n[n]
                        else:
                            weight = kpt.weight - kpt.f_n[n]
                        self.add_wf_to_ldos(n, u, weight)

    def add_wf_to_ldos(self, n, u, weight=None):
        """Add the wf with given kpoint and spin to the ldos"""
        kpt = self.calc.wfs.kpt_u[u]
        psi = kpt.psit_nG[n]
        w = weight
        if w is None:
            w = kpt.weight
# print "w=", w, kpt.weight
        self.ldos += w * (psi * np.conj(psi)).real

    def write_3D(self, bias, file, filetype=None):
        """Write the density as a 3D file.

        Units: [e/A^3]"""
        self.calculate_ldos(bias)
        self.calc.wfs.kd.comm.sum(self.ldos)
        ldos = self.gd.collect(self.ldos)
# print "write: integrated =", self.gd.integrate(self.ldos)

        if mpi.rank != 0:
            return

        if filetype is None:
            # estimate file type from name ending
            filetype = file.split('.')[-1]
        filetype = filetype.lower()

        if filetype == 'cube':
            with open(file, 'w') as fd:
                write_cube(fd, self.calc.get_atoms(), ldos / Bohr ** 3)
        else:
            raise NotImplementedError('unknown file type "' + filetype + '"')

    def read_3D(self, file, filetype=None):
        """Read the density from a 3D file"""

        if filetype is None:
            # estimate file type from name ending
            filetype = file.split('.')[-1]

        if filetype == 'cube':
            data, atoms = read_cube_data(file)
            self.cell = atoms.cell.array

            pbc_c = [True, True, True]
            N_c = np.array(data.shape)
            for c in range(3):
                if N_c[c] % 2 == 1:
                    pbc_c[c] = False
                    N_c[c] += 1
            self.gd = GridDescriptor(N_c, self.cell.diagonal() / Bohr, pbc_c)
            self.offset_c = [int(not a) for a in self.gd.pbc_c]

        else:
            raise NotImplementedError('unknown file type "' + filetype + '"')

        self.file = file
        self.ldos = np.array(data * Bohr ** 3)
# print "read: integrated =", self.gd.integrate(self.ldos)

    def current_to_density(self, current):
        """The connection between density n and current I

        n [e/Angstrom^3] = 0.0002 sqrt(I [nA])

        as given in Hofer et al., RevModPhys 75 (2003) 1287
        """
        return 0.0002 * sqrt(current)

    def linescan(self, bias, current, p1, p2, npoints=50, z0=None):
        """Line scan for a given current [nA]"""
        return STM.linescan(self, bias,
                            self.current_to_density(current),
                            p1, p2, npoints, z0)

    def density_to_current(self, density):
        return 5000. * density ** 2

    def scan_const_current(self, current, bias):
        """Get the height image for constant current I [nA].
        """
        return self.scan_const_density(self.current_to_density(current),
                                       bias)

    def scan_const_density(self, density, bias):
        """Get the height image for constant density [e/Angstrom^3].
        """
        x, y, heights = self.scan(bias, density)
        self.heights = heights / Bohr

        return self.heights

    def write(self, file=None):
        """Write STM data to a file in gnuplot readable tyle."""

        if mpi.rank != 0:
            return

        xvals, yvals, heights = self.pylab_contour()
        nx, ny = heights.shape[:2]

        if file is None:
            n, k, s = self.bias
            fname = 'stm_n%dk%ds%d.dat' % (n, k, s)
        else:
            fname = file
        
        with open(fname, 'w') as fd:

            import datetime
            print('#', datetime.datetime.now().ctime(), file=fd)
            print('# Simulated STM picture', file=fd)
            if hasattr(self, 'file'):
                print('# density read from', self.file, file=fd)
            else:
                if self.is_wf:
                    print('# pseudo-wf n=%d k=%d s=%d' % tuple(self.bias),
                          file=fd)
                else:
                    print('# bias=', self.bias, '[eV]', file=fd)
            print('#', file=fd)
            print('# density=', self.density, '[e/Angstrom^3]',
                  end=' ', file=fd)
            print(
                '(current=', self.density_to_current(self.density), '[nA])',
                file=fd)
            print('# x[Angs.]   y[Angs.]     h[Angs.] (-1 is not found)',
                  file=fd)
            for i in range(nx):
                for j in range(ny):
                    if heights[i, j] == -1:
                        height = -1
                    else:
                        height = heights[i, j] * Bohr
                    print('%10g %10g %12g' % (yvals[j], xvals[i], height),
                          file=fd)
                print(file=fd)

    def pylab_contour(self):
        """Return the countour to be plotted using pylab."""

        nx, ny = self.heights.shape[:2]
        h_c = np.array([np.linalg.norm(self.gd.h_cv[c])
                        for c in range(3)]) * Bohr
        # the lowest point is not stored for non-periodic BCs
        xvals = [(i + self.offset_c[0]) * h_c[0] for i in range(nx)]
        yvals = [(i + self.offset_c[1]) * h_c[1] for i in range(ny)]
        heights = self.heights * Bohr

        # pylab interprets heights[y_i][x_i]
        return np.array(xvals), np.array(yvals), heights.swapaxes(0, 1)



from ase.spacegroup import crystal
from gpaw import GPAW
from gpaw import PW
from gpaw.test import equal


def test_symmetry_fractional_translations_big():
    'cristobalite'
    # no. 92 - tetragonal

    a = 5.0833674
    c = 7.0984738
    p0 = (0.2939118, 0.2939118, 0.0)
    p1 = (0.2412656, 0.0931314, 0.1739217)

    atoms = crystal(['Si', 'O'], basis=[p0, p1],
                    spacegroup=92, cellpar=[a, a, c, 90, 90, 90])

    # with fractional translations
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 2),
                nbands=40,
                symmetry={'symmorphic': False},
                gpts=(24, 24, 32),
                eigensolver='rmm-diis')

    atoms.calc = calc
    energy_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 3
    assert len(calc.wfs.kd.symmetry.op_scc) == 8

    # without fractional translations
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 2),
                nbands=40,
                gpts=(24, 24, 32),
                eigensolver='rmm-diis')

    atoms.calc = calc
    energy_no_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 6
    assert len(calc.wfs.kd.symmetry.op_scc) == 2

    equal(energy_fractrans, energy_no_fractrans, 1e-7)



import pytest
from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw.test import equal


@pytest.mark.gllb
@pytest.mark.libxc
def test_gllb_spin():
    for spin in [False, True]:
        a = 3.56
        atoms = bulk('C', 'diamond', a=a)
        calc = GPAW(kpts=(3, 3, 3),
                    xc='GLLBSC',
                    spinpol=spin,
                    nbands=8,
                    convergence={'bands': 6, 'density': 1e-6},
                    occupations=FermiDirac(width=0.005))
        atoms.calc = calc
        atoms.get_potential_energy()
        resp = calc.hamiltonian.xc.response
        # Eks is the Kohn-Sham gap and Dxc is the derivative discontinuity
        if spin:
            homoa, lumoa = calc.get_homo_lumo(spin=0)
            homob, lumob = calc.get_homo_lumo(spin=1)
            dxc_pot = resp.calculate_discontinuity_potential((homoa, homob),
                                                             (lumoa, lumob))
            Eksa, Dxca = resp.calculate_discontinuity(dxc_pot, spin=0)
            Eksb, Dxcb = resp.calculate_discontinuity(dxc_pot, spin=1)
            Gapa = Eksa + Dxca
            Gapb = Eksb + Dxcb
            print("GAP", spin, Gapa, Gapb)
        else:
            homo, lumo = calc.get_homo_lumo()
            dxc_pot = resp.calculate_discontinuity_potential(homo, lumo)
            Eks, Dxc = resp.calculate_discontinuity(dxc_pot)
            Gap = Eks + Dxc
            print("GAP", spin, Gap)

    equal(Gapa, Gapb, 1e-4)
    equal(Gapa, Gap, 1e-4)



from gpaw.xc import XC
from gpaw.utilities import unpack2, pack
from gpaw.hybrids.paw import pawexxvv
import numpy as np


def RI(name, **args):
    if name == 'HSE06WIP':
        return RIXC(name, exx_fraction=0.25,
                    omega=0.11, localxc='HYB_GGA_XC_HSE06')
    raise NotImplementedError(f'RI backend does not support '
                              f'functional called: {name}.')


EXXCW_NONE_ERROR = 'The setup does not contain erfc-screened exchange '\
                   'core-core energy for given omega.'


class RIXC:
    orbital_dependent = False  # SIC!
    type = 'ri'
    # orbital_dependent_lcao = True
    # type = 'auxhybrid'

    def __init__(self, name, *, exx_fraction=None, omega=None, localxc=None):
        self.name = name
        self.exx_fraction = exx_fraction
        self.omega = omega
        self.localxc = XC(localxc)

    def initialize(self, density, hamiltonian, wfs):
        self.density = density
        self.ecc = 0
        for setup in wfs.setups:
            if self.omega is not None:
                if setup.ExxC_w is None:
                    raise RuntimeError(EXXCW_NONE_ERROR)
                self.ecc += setup.ExxC_w[self.omega] * self.exx_fraction
            else:
                self.ecc += setup.ExxC * self.exx_fraction
        # self.ri_algorithm.initialize(density, hamiltonian, wfs)

    def get_setup_name(self):
        return 'PBE'

    def set_grid_descriptor(self, gd):
        pass

    def set_positions(self, spos_ac):
        pass

    def get_description(self):
        return f'Resolution of identity (RI) for {self.name} functional.\n'\
               f'   EXX fraction: {self.exx_fraction}\n'\
               f'   Kernel      : erfc(wr)/r\n'\
               f'   omega (w)   : {self.omega}\n'

    def calculate(self, gd, nt_sr, vt_sr):
        """
             Calculate the semi-local potential of the hybrid.
        """
        # if self.use_lda:
        #     return self.ldaxc.calculate(gd, nt_sr, vt_sr)
        energy = self.ecc  # + self.evv + self.evc
        energy += self.localxc.calculate(gd, nt_sr, vt_sr)
        return energy

    def calculate_paw_correction(self, setup, D_sp, dH_sp=None, a=None):
        E = self.localxc.calculate_paw_correction(setup, D_sp, dH_sp, a=a)

        for s, (dH_p, D_p) in enumerate(zip(dH_sp, D_sp)):
            D_ii = unpack2(D_p)
            V_ii = pawexxvv(setup.M_wpp[self.omega], D_ii)  # *(dens.nspins/2)
            dH_p += pack(V_ii)

            E += np.sum(V_ii.ravel() * D_ii.ravel())  # * prefactor

            dH_p -= self.density.setups[a].X_p * self.exx_fraction
            E -= self.exx_fraction * np.dot(D_p, self.density.setups[a].X_p)

        return E

    def get_kinetic_energy_correction(self):
        return 0  # self.ekin

    def summary(self, log):
        # Take no changes
        # if self.use_lda:
        #     raise ValueError('Error: Due to an
        # internal error, LDA was used thorough the calculation.')

        log(self.get_description())



""" Tests extrapolation to infinite energy cutoff + block parallelization.
It takes ~10 s on one core"""

import pytest
from gpaw.test import equal
from gpaw.response.g0w0 import G0W0


@pytest.mark.response
def test_response_gw_hBN_extrapolate(in_tmp_dir, scalapack, gpw_files,
                                     needs_ase_master):
    gw = G0W0(gpw_files['hbn_pw_wfs'],
              'gw-hBN',
              ecut=50,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1},
              eta=0.2,
              truncation='2D',
              q0_correction=True,
              kpts=[0],
              bands=(3, 5),
              ecut_extrapolation=[20, 25, 30],
              nblocksmax=True)

    e_qp = gw.calculate()['qp'][0, 0]

    ev = -1.7321
    ec = 3.1999
    equal(e_qp[0], ev, 0.01)
    equal(e_qp[1], ec, 0.01)



from __future__ import annotations
from math import nan
from operator import attrgetter
from pathlib import Path
from typing import (TYPE_CHECKING, Callable, Dict, Iterable, Iterator, List,
                    Optional, Tuple)

import numpy as np
from ase.units import Bohr, Ha, alpha

from gpaw.band_descriptor import BandDescriptor
from gpaw.grid_descriptor import GridDescriptor
from gpaw.kpoint import KPoint
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import broadcast_array, serial_comm
from gpaw.occupations import OccupationNumberCalculator, ParallelLayout
from gpaw.projections import Projections
from gpaw.setup import Setup
from gpaw.typing import Array1D, Array2D, Array3D, Array4D, ArrayND
from gpaw.utilities.ibz2bz import construct_symmetry_operators
from gpaw.utilities.partition import AtomPartition

if TYPE_CHECKING:
    from gpaw.calculator import GPAW  # noqa
    from gpaw.new.ase_interface import ASECalculator

_L_vlmm: List[List[np.ndarray]] = []  # see get_L_vlmm() below


class WaveFunction:
    def __init__(self,
                 eigenvalues: Array1D,
                 projections: Projections,
                 bz_index: int = None):
        self.eig_m = eigenvalues
        self.projections = projections
        self.spin_projection_mv: Optional[Array2D] = None
        self.v_mn: Optional[Array2D] = None
        self.f_m = np.empty_like(self.eig_m)
        self.f_m[:] = nan
        self.bz_index = bz_index

    def transform(self,
                  kd: KPointDescriptor,
                  setups: List[Setup],
                  spos_ac: Array2D,
                  bz_index: int) -> 'WaveFunction':
        """Transforms PAW projections from IBZ to BZ k-point."""
        a_a, U_aii, time_rev = construct_symmetry_operators(
            kd, setups, spos_ac, bz_index)

        projections = self.projections.new()

        if projections.atom_partition.comm.rank == 0:
            a = 0
            for b, U_ii in zip(a_a, U_aii):
                P_msi = self.projections[b].dot(U_ii)
                if time_rev:
                    P_msi = P_msi.conj()
                projections[a][:] = P_msi
                a += 1
        else:
            assert len(projections.indices) == 0

        return WaveFunction(self.eig_m.copy(), projections, bz_index)

    def redistribute_atoms(self,
                           atom_partition: AtomPartition
                           ) -> 'WaveFunction':
        projections = self.projections.redist(atom_partition)
        return WaveFunction(self.eig_m.copy(), projections, self.bz_index)

    def add_soc(self,
                dVL_avii: Dict[int, Array3D],
                s_vss: List[Array2D],
                C_ss: Array2D) -> None:
        """Evaluate H in a basis of S_z eigenstates."""
        if self.projections.bcomm.rank > 0:
            return

        M = self.projections.nbands
        H_mm = np.zeros((M, M), complex)
        for a, dVL_vii in dVL_avii.items():
            ni = dVL_vii.shape[1]
            H_ssii = np.zeros((2, 2, ni, ni), complex)
            H_ssii[0, 0] = dVL_vii[2]
            H_ssii[0, 1] = dVL_vii[0] - 1.0j * dVL_vii[1]
            H_ssii[1, 0] = dVL_vii[0] + 1.0j * dVL_vii[1]
            H_ssii[1, 1] = -dVL_vii[2]

            # Tranform to theta, phi basis
            H_ssii = np.tensordot(C_ss, H_ssii, (0, 1))
            H_ssii = np.tensordot(C_ss.T.conj(), H_ssii, (1, 1))
            H_ssii *= Ha

            P_msi = self.projections[a]
            for s1 in range(2):
                for s2 in range(2):
                    H_ii = H_ssii[s1, s2]
                    P1_mi = P_msi[:, s1]
                    P2_mi = P_msi[:, s2]
                    H_mm += np.dot(np.dot(P1_mi.conj(), H_ii), P2_mi.T)

        domain_comm = self.projections.atom_partition.comm
        domain_comm.sum(H_mm, 0)
        if domain_comm.rank == 0:
            H_mm += np.diag(self.eig_m)
            self.eig_m, v_nm = np.linalg.eigh(H_mm)
        else:
            v_nm = np.empty((M, M), complex)

        domain_comm.broadcast(v_nm, 0)

        P_mI = self.projections.matrix.array
        P_mI[:] = v_nm.T.copy().dot(P_mI)

        sx_m = []
        sy_m = []
        sz_m = []

        v_msn = v_nm.copy().reshape((M // 2, 2, M)).T.copy()
        for v_sn in v_msn:
            sx_m.append(np.trace(v_sn.T.conj().dot(s_vss[0]).dot(v_sn)))
            sy_m.append(np.trace(v_sn.T.conj().dot(s_vss[1]).dot(v_sn)))
            sz_m.append(np.trace(v_sn.T.conj().dot(s_vss[2]).dot(v_sn)))

        self.spin_projection_mv = np.array([sx_m, sy_m, sz_m]).real.T.copy()
        self.v_mn = v_nm.T

    def wavefunctions(self, calc, periodic=True):
        kd = calc.wfs.kd
        assert kd.nibzkpts == kd.nbzkpts

        # For spinors the number of bands is doubled and a
        # spin dimension is added
        Ns = calc.wfs.nspins
        Nm, Nn = self.v_mn.shape

        if calc.wfs.collinear:
            u_snR = [[calc.wfs.get_wave_function_array(n, self.bz_index, s,
                                                       periodic=periodic)
                      for n in range(Nn // 2)]
                     for s in range(Ns)]
            u_msR = np.empty((Nm, 2) + u_snR[0][0].shape, complex)
            np.einsum('mn, nabc -> mabc', self.v_mn[:, ::2], u_snR[0],
                      out=u_msR[:, 0])
            np.einsum('mn, nabc -> mabc', self.v_mn[:, 1::2], u_snR[-1],
                      out=u_msR[:, 1])
        else:
            u_nsR = np.array(
                [calc.wfs.get_wave_function_array(n, self.bz_index, 0,
                                                  periodic=periodic)
                 for n in range(Nn)])
            u_msR = np.einsum('mn, nsxyz -> msxyz',
                              self.v_mn, u_nsR)

        return u_msR

    @property
    def P_amj(self):
        M = self.projections.nbands
        return {
            a: P_msi.transpose((0, 2, 1)).copy().reshape((M, -1))
            for a, P_msi in self.projections.items()}

    def pdos_weights(self,
                     a: int,
                     indices: List[int]
                     ) -> Array3D:
        """PDOS weights."""
        dos_ms = np.zeros((self.projections.nbands, 2))

        P_amsi = self.projections

        if a in P_amsi:
            dos_ms[:, :] = (abs(P_amsi[a][:, :, indices])**2).sum(2)

        return dos_ms


class BZWaveFunctions:
    """Container for eigenvalues and PAW projections (all of BZ)."""
    def __init__(self,
                 kd: KPointDescriptor,
                 wfs: Dict[int, WaveFunction],
                 occ: Optional[OccupationNumberCalculator],
                 nelectrons: float):
        self.wfs = wfs
        self.occ = occ
        self.nelectrons = nelectrons

        self.nbzkpts = kd.nbzkpts

        # Initialize ranks:
        self.ranks = np.zeros(kd.nbzkpts, int)
        for k in wfs:
            self.ranks[k] = kd.comm.rank
        kd.comm.sum(self.ranks)

        wf = next(iter(wfs.values()))  # get the first WaveFunction object

        self.shape = (kd.nbzkpts, wf.projections.nbands)
        self.domain_comm = wf.projections.atom_partition.comm
        self.bcomm = wf.projections.bcomm
        self.kpt_comm = kd.comm

        self.fermi_level = self._calculate_occ_numbers_and_fermi_level()

        self.size = kd.N_c
        self.bz2ibz_map = np.arange(self.nbzkpts)

    def weights(self):
        return np.zeros(len(self)) + 1 / self.nbzkpts

    def _calculate_occ_numbers_and_fermi_level(self) -> float:
        if self.occ is not None:
            eig_im = [wf.eig_m for wf in self]
            weight = 1.0 / self.nbzkpts
            weight_i = [weight] * len(eig_im)

            f_im, (fermi_level,), _ = self.occ.calculate(
                self.nelectrons,
                eig_im,
                weight_i)
            for wf, f_m in zip(self, f_im):
                wf.f_m[:] = f_m
        else:
            fermi_level = 0.0

        if self.domain_comm.rank == 0 and self.bcomm.rank == 0:
            fermi_level = self.bcomm.sum(fermi_level)
        fermi_level = self.domain_comm.sum(fermi_level)

        return fermi_level

    def calculate_band_energy(self) -> float:
        """Calculate sum over occupied eigenvalues."""
        if self.domain_comm.rank == 0 and self.bcomm.rank == 0:
            weight = 1.0 / self.nbzkpts
            e_band = sum(wf.eig_m.dot(wf.f_m) for wf in self) * weight
            e_band = self.kpt_comm.sum(e_band)
        else:
            e_band = 0.0

        if self.domain_comm.rank == 0:
            e_band = self.bcomm.sum(e_band)
        e_band = self.domain_comm.sum(e_band)
        return e_band

    def __iter__(self):
        for bz_index in sorted(self.wfs):
            yield self[bz_index]

    def __getitem__(self, bz_index):
        return self.wfs[bz_index]

    def __len__(self):
        return len(self.wfs)

    def eigenvalues(self,
                    broadcast: bool = True
                    ) -> Array2D:
        """Eigenvalues in eV for the whole BZ."""
        return self._collect(attrgetter('eig_m'), broadcast=broadcast)

    def occupation_numbers(self,
                           broadcast: bool = True
                           ) -> Array2D:
        """Occupation numbers for the whole BZ."""
        return self._collect(attrgetter('f_m'), broadcast=broadcast)

    def eigenvectors(self,
                     broadcast: bool = True
                     ) -> Array4D:
        """Eigenvectors for the whole BZ."""
        nbands = self.shape[1]
        assert nbands % 2 == 0
        return self._collect(attrgetter('v_mn'), (nbands,), complex,
                             broadcast=broadcast)

    def spin_projections(self,
                         broadcast: bool = True
                         ) -> Array3D:
        """Spin projections for the whole BZ."""
        return self._collect(attrgetter('spin_projection_mv'), (3,),
                             broadcast=broadcast)

    def pdos_weights(self,
                     a: int,
                     indices: List[int],
                     broadcast: bool = True
                     ) -> Array4D:
        """Projections for PDOS.

        Returns (nbzkpts, nbands, 2)-shaped ndarray
        of the square of absolute value of the projections.
        """
        def func(wf):
            return wf.pdos_weights(a, indices)

        return self._collect(func,
                             (2,),
                             broadcast=broadcast,
                             sum_over_domain=True)

    def _collect(self,
                 func: Callable[[WaveFunction], ArrayND],
                 shape: Tuple[int, ...] = None,
                 dtype=float,
                 broadcast: bool = True,
                 sum_over_domain: bool = False) -> ArrayND:
        """Helper method for collecting (and broadcasting) ndarrays."""

        total_shape = self.shape + (shape or ())

        if broadcast:
            array_kmx = self._collect(func,
                                      shape,
                                      dtype,
                                      False,
                                      sum_over_domain)
            if array_kmx.ndim == 0:
                array_kmx = np.empty(total_shape, dtype)
            return broadcast_array(array_kmx,
                                   self.kpt_comm, self.bcomm, self.domain_comm)

        if self.bcomm.rank != 0:
            return np.empty(shape=())

        if not sum_over_domain and self.domain_comm.rank != 0:
            return np.empty(shape=())

        comm = self.kpt_comm
        if comm.rank == 0:
            array_kmx = np.empty(total_shape, dtype)
            for k, rank in enumerate(self.ranks):
                if rank == 0:
                    array_kmx[k] = func(self[k])
                else:
                    comm.receive(array_kmx[k], rank)
            if sum_over_domain:
                self.domain_comm.sum(array_kmx)
            if self.domain_comm.rank == 0:
                return array_kmx
            else:
                return np.empty(shape=())

        for k, rank in enumerate(self.ranks):
            if rank == comm.rank:
                comm.send(func(self[k]), 0)

        return np.empty(shape=())


def soc_eigenstates_raw(ibzwfs: Iterable[Tuple[int, WaveFunction]],
                        dVL_avii: Dict[int, Array3D],
                        kd, spos_ac, setups, atom_partition,
                        theta: float = 0.0,
                        phi: float = 0.0) -> Dict[int, WaveFunction]:

    theta *= np.pi / 180
    phi *= np.pi / 180

    # Hamiltonian with SO in KS basis
    # The even indices in H_mm are spin up along \hat n defined by \theta, phi
    # Basis change matrix for constructing Pauli matrices in \theta,\phi basis:
    #     \sigma_i^n = C^\dag\sigma_i C
    C_ss = np.array([[np.cos(theta / 2) * np.exp(-1.0j * phi / 2),
                      -np.sin(theta / 2) * np.exp(-1.0j * phi / 2)],
                     [np.sin(theta / 2) * np.exp(1.0j * phi / 2),
                      np.cos(theta / 2) * np.exp(1.0j * phi / 2)]])

    sx_ss = np.array([[0, 1], [1, 0]], complex)
    sy_ss = np.array([[0, -1.0j], [1.0j, 0]], complex)
    sz_ss = np.array([[1, 0], [0, -1]], complex)
    s_vss = [C_ss.T.conj() @ sx_ss @ C_ss,
             C_ss.T.conj() @ sy_ss @ C_ss,
             C_ss.T.conj() @ sz_ss @ C_ss]

    bzwfs = {}
    for ibz_index, ibzwf in ibzwfs:
        for bz_index in np.nonzero(kd.bz2ibz_k == ibz_index)[0]:
            bzwf = ibzwf.transform(kd, setups, spos_ac, bz_index)

            # Redistribute to match dVL_avii:
            bzwf = bzwf.redistribute_atoms(atom_partition)

            bzwf.add_soc(dVL_avii, s_vss, C_ss)
            bzwfs[bz_index] = bzwf

    return bzwfs


def extract_ibz_wave_functions(kpt_qs: List[List[KPoint]],
                               bd: BandDescriptor,
                               gd: GridDescriptor,
                               n1: int,
                               n2: int,
                               eigenvalues: Array3D = None
                               ) -> Iterator[Tuple[int, WaveFunction]]:
    """Yield tuples of IBZ-index and wave functions.

    All atoms and bands will be on rank == 0 of gd.comm and bd.comm
    respectively.  This makes slicing the bands (from n1 to n2-1)
    and symmetry operations on the projections easier.
    """

    nproj_a = kpt_qs[0][0].projections.nproj_a

    collinear = kpt_qs[0][0].s is not None

    nbands = n2 - n1
    if collinear:
        nbands *= 2

    # All atoms on rank-0:
    atom_partition = AtomPartition(gd.comm, np.zeros_like(nproj_a))

    # All bands on rank-0 (nrow * ncol = 1 * 1):
    bdist = (bd.comm, 1, 1)

    for kpt_s in kpt_qs:
        # Collect bands and atoms to bd.comm.rank == 0 and gd.comm.rank == 0:
        if eigenvalues is None:
            eig_sn = [bd.collect(kpt.eps_n) for kpt in kpt_s]
        P_snI = [kpt.projections.collect() for kpt in kpt_s]

        projections = Projections(
            nbands=nbands,
            nproj_a=nproj_a,
            atom_partition=atom_partition,
            bdist=bdist,
            collinear=False)

        if bd.comm.rank == 0 and gd.comm.rank == 0:
            P1_nI = P_snI[0]
            P2_nI = P_snI[-1]
            assert P1_nI is not None
            assert P2_nI is not None
            if collinear:
                eig_m = np.empty((n2 - n1) * 2)
                if eigenvalues is None:
                    eig_m[::2] = eig_sn[0][n1:n2] * Ha
                    eig_m[1::2] = eig_sn[-1][n1:n2] * Ha
                else:
                    for s in range(2):
                        eig_m[s::2] = eigenvalues[-s, kpt_s[-s].k]
                projections.array[:] = 0.0
                projections.array[::2, 0] = P1_nI[n1:n2]
                projections.array[1::2, 1] = P2_nI[n1:n2]
            else:
                eig_m = eig_sn[0][n1:n2] * Ha
                projections.matrix.array[:] = P1_nI[n1:n2]
        else:
            eig_m = np.empty(0)

        ibz_index = kpt_s[0].k

        yield ibz_index, WaveFunction(eig_m, projections)


def soc_eigenstates(calc: ASECalculator | GPAW | str | Path,
                    n1: int = None,
                    n2: int = None,
                    scale: float = 1.0,
                    theta: float = 0.0,  # degrees
                    phi: float = 0.0,  # degrees
                    eigenvalues: Array3D = None,  # eV
                    occcalc: OccupationNumberCalculator = None,
                    projected: bool = False
                    ) -> BZWaveFunctions:
    """Calculate SOC eigenstates.

    Parameters:
        calc: Calculator
            GPAW calculator or path to gpw-file.
        n1, n2: int
            Range of bands to include (n1 <= n < n2).  Default is all
            bands available.
        scale: float
            Scale the spinorbit coupling by this amount.
        theta: float
            Angle in degrees.
        phi: float
            Angle in degrees.
        eigenvalues: ndarray
            Optionally use these eigenvalues instead for those from *calc*.
            The shape must be: (nspins, nibzkpts, n2 - n1).  Units: eV.
        occcalc:
            Occupation-number calculator.  By default, the one from *calc*
            will be used.

    Returns a BZWaveFunctions object covering the whole BZ.
    """

    from gpaw.calculator import GPAW  # noqa

    if isinstance(calc, (str, Path)):
        calc = GPAW(calc)

    n1 = n1 or 0
    n2 = n2 or 0
    if n2 <= 0:
        if eigenvalues is None:
            nbands = calc.get_number_of_bands()
        else:
            nbands = eigenvalues.shape[2]
        n2 += nbands

    # <phi_i|dV_adr / r * L_v|phi_j>
    dVL_avii = {a: soc(calc.wfs.setups[a],
                       calc.hamiltonian.xc, D_sp) * scale
                for a, D_sp in calc.density.D_asp.items()}

    if projected:
        dVL_avii = {a: projected_soc(dVL_vii, theta=theta, phi=phi)
                    for a, dVL_vii in dVL_avii.items()}

    kd = calc.wfs.kd
    bd = calc.wfs.bd
    gd = calc.wfs.gd
    spos_ac = calc.spos_ac
    setups = calc.wfs.setups
    atom_partition = calc.density.atom_partition

    if eigenvalues is not None:
        assert eigenvalues.shape == (kd.nspins, kd.nibzkpts, n2 - n1)

    ibzwfs = extract_ibz_wave_functions(calc.wfs.kpt_qs,
                                        bd, gd, n1, n2, eigenvalues)

    bzwfs = soc_eigenstates_raw(ibzwfs,
                                dVL_avii,
                                kd, spos_ac, setups, atom_partition,
                                theta, phi)

    if bd.comm.rank == 0 and gd.comm.rank == 0:
        parallel_layout = ParallelLayout(BandDescriptor(1),
                                         kd.comm,
                                         serial_comm)
        occcalc = occcalc or calc.wfs.occupations
        occcalc = occcalc.copy(bz2ibzmap=list(range(kd.nbzkpts)),
                               parallel_layout=parallel_layout)
    else:
        occcalc = None

    return BZWaveFunctions(kd, bzwfs, occcalc, calc.wfs.nvalence)


def soc(a: Setup, xc, D_sp: Array2D) -> Array3D:
    """<phi_i|dV_adr / r * L_v|phi_j>"""
    v_g = get_radial_potential(a, xc, D_sp)
    Ng = len(v_g)
    phi_jg = a.data.phi_jg

    Lx_lmm, Ly_lmm, Lz_lmm = get_L_vlmm()

    dVL_vii = np.zeros((3, a.ni, a.ni), complex)
    N1 = 0
    for j1, l1 in enumerate(a.l_j):
        Nm = 2 * l1 + 1
        N2 = 0
        for j2, l2 in enumerate(a.l_j):
            if l1 == l2:
                f_g = phi_jg[j1][:Ng] * v_g * phi_jg[j2][:Ng]
                c = a.xc_correction.rgd.integrate(f_g) / (4 * np.pi)
                dVL_vii[0, N1:N1 + Nm, N2:N2 + Nm] = c * Lx_lmm[l1]
                dVL_vii[1, N1:N1 + Nm, N2:N2 + Nm] = c * Ly_lmm[l1]
                dVL_vii[2, N1:N1 + Nm, N2:N2 + Nm] = c * Lz_lmm[l1]
            else:
                pass
            N2 += 2 * l2 + 1
        N1 += Nm
    return dVL_vii * alpha**2 / 4.0


def projected_soc(dVL_vii: Array3D,
                  theta: float = 0,
                  phi: float = 0) -> Array3D:
    """
    Optional Args:
        theta (float): The angle from z-axis in degrees
        phi (float): The angle from x-axis in degrees
    """
    theta *= np.pi / 180
    phi *= np.pi / 180
    n_v = np.array([np.sin(theta) * np.cos(phi),
                    np.sin(theta) * np.sin(phi),
                    np.cos(theta)])
    dVL_vii = (np.dot(dVL_vii.T, n_v)[:, :, np.newaxis] * n_v).T
    return dVL_vii


def get_radial_potential(a: Setup, xc, D_sp: Array2D) -> Array1D:
    """Calculates (dV/dr)/r for the effective potential.
    Below, f_g denotes dV/dr = minus the radial force"""

    rgd = a.xc_correction.rgd
    r_g = rgd.r_g.copy()
    r_g[0] = 1.0e-12
    dr_g = rgd.dr_g

    B_pq = a.xc_correction.B_pqL[:, :, 0]
    n_qg = a.xc_correction.n_qg
    D_sq = np.dot(D_sp, B_pq)
    n_sg = np.dot(D_sq, n_qg) / (4 * np.pi)**0.5
    Ns = len(D_sp)
    if Ns == 4:
        Ns = 1
    n_sg[:Ns] += a.xc_correction.nc_g / Ns

    # Coulomb force from nucleus
    fc_g = a.Z / r_g**2

    # Hartree force
    rho_g = 4 * np.pi * r_g**2 * dr_g * np.sum(n_sg[:Ns], axis=0)
    fh_g = -np.array([np.sum(rho_g[:ig]) for ig in range(len(r_g))]) / r_g**2

    f_g = fc_g + fh_g

    # xc force
    if xc.type != 'GLLB':
        v_sg = np.zeros_like(n_sg)
        xc.calculate_spherical(rgd, n_sg, v_sg)
        fxc_g = np.mean([rgd.derivative(v_g) for v_g in v_sg[:Ns]],
                        axis=0)
        f_g += fxc_g

    return f_g / r_g


def get_anisotropy(calc, theta=0.0, phi=0.0, nbands=0, width=None):
    """Calculates the sum of occupied spinorbit eigenvalues.

    Returns the result relative to the sum of eigenvalues without
    spinorbit coupling.
    """
    raise RuntimeError('Please use BZWaveFunctions.calculate_band_energy() '
                       'instead.')


def get_magnetic_moments(calc, theta=0.0, phi=0.0, nbands=None, width=None):
    """Calculates the magnetic moments inside all PAW spheres"""

    raise RuntimeError(
        'This function has no tests.  It is very likely that it no longer '
        'works correctly after merging !677.')

    from gpaw.utilities import unpack

    if nbands is None:
        nbands = calc.get_number_of_bands()
    Nk = len(calc.get_ibz_k_points())

    C_ss = np.array([[np.cos(theta / 2) * np.exp(-1.0j * phi / 2),
                      -np.sin(theta / 2) * np.exp(-1.0j * phi / 2)],
                     [np.sin(theta / 2) * np.exp(1.0j * phi / 2),
                      np.cos(theta / 2) * np.exp(1.0j * phi / 2)]])
    sx_ss = np.array([[0, 1], [1, 0]], complex)
    sy_ss = np.array([[0, -1.0j], [1.0j, 0]], complex)
    sz_ss = np.array([[1, 0], [0, -1]], complex)
    sx_ss = C_ss.T.conj().dot(sx_ss).dot(C_ss)
    sy_ss = C_ss.T.conj().dot(sy_ss).dot(C_ss)
    sz_ss = C_ss.T.conj().dot(sz_ss).dot(C_ss)

    states = soc_eigenstates(calc,
                             theta=theta,
                             phi=phi,
                             return_wfs=True,
                             bands=range(nbands))
    e_km = states['eigenvalues']
    v_knm = states['eigenstates']

    from gpaw.occupations import occupation_numbers
    if width is None:
        assert calc.wfs.occupations.name == 'fermi-dirac'
        width = calc.wfs.occupations._width
    if width == 0.0:
        width = 1.e-6
    weight_k = calc.get_k_point_weights() / 2
    ne = calc.wfs.setups.nvalence - calc.density.charge
    f_km = occupation_numbers({'name': 'fermi-dirac', 'width': width},
                              e_km[np.newaxis],
                              weight_k=weight_k,
                              nelectrons=ne)[0][0]

    m_v = np.zeros(3, complex)
    for ik in range(Nk):
        ut0_nG = np.array([calc.wfs.get_wave_function_array(n, ik, 0)
                           for n in range(nbands)])
        ut1_nG = np.array([calc.wfs.get_wave_function_array(n, ik, 1)
                           for n in range(nbands)])
        mocc = np.where(f_km[ik] * Nk - 1.0e-6 < 0.0)[0][0]
        for m in range(mocc + 1):
            f = f_km[ik, m]
            ut0_G = np.dot(v_knm[ik][::2, m], np.swapaxes(ut0_nG, 0, 2))
            ut1_G = np.dot(v_knm[ik][1::2, m], np.swapaxes(ut1_nG, 0, 2))
            ut_sG = np.array([ut0_G, ut1_G])

            mx_G = np.zeros(np.shape(ut0_G), complex)
            my_G = np.zeros(np.shape(ut0_G), complex)
            mz_G = np.zeros(np.shape(ut0_G), complex)
            for s1 in range(2):
                for s2 in range(2):
                    mx_G += ut_sG[s1].conj() * sx_ss[s1, s2] * ut_sG[s2]
                    my_G += ut_sG[s1].conj() * sy_ss[s1, s2] * ut_sG[s2]
                    mz_G += ut_sG[s1].conj() * sz_ss[s1, s2] * ut_sG[s2]
            m_v += calc.wfs.gd.integrate(np.array([mx_G, my_G, mz_G])) * f

    m_av = []
    for a in range(len(calc.atoms)):
        N0_p = calc.density.setups[a].N0_p.copy()
        N0_ij = unpack(N0_p)
        Dx_ij = np.zeros_like(N0_ij, complex)
        Dy_ij = np.zeros_like(N0_ij, complex)
        Dz_ij = np.zeros_like(N0_ij, complex)
        Delta_p = calc.density.setups[a].Delta_pL[:, 0].copy()
        Delta_ij = unpack(Delta_p)
        for ik in range(Nk):
            P_ami = ...  # get_spinorbit_projections(calc, ik, v_knm[ik])
            P_smi = np.array([P_ami[a][:, ::2], P_ami[a][:, 1::2]])
            P_smi = np.dot(C_ss, np.swapaxes(P_smi, 0, 1))

            P0_mi = P_smi[0]
            P1_mi = P_smi[1]
            f_mm = np.diag(f_km[ik])

            Dx_ij += P0_mi.conj().T.dot(f_mm).dot(P1_mi)
            Dx_ij += P1_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dy_ij -= 1.0j * P0_mi.conj().T.dot(f_mm).dot(P1_mi)
            Dy_ij += 1.0j * P1_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dz_ij += P0_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dz_ij -= P1_mi.conj().T.dot(f_mm).dot(P1_mi)

        mx = np.sum(N0_ij * Dx_ij).real
        my = np.sum(N0_ij * Dy_ij).real
        mz = np.sum(N0_ij * Dz_ij).real

        m_av.append([mx, my, mz])
        m_v[0] += np.sum(Delta_ij * Dx_ij) * (4 * np.pi)**0.5
        m_v[1] += np.sum(Delta_ij * Dy_ij) * (4 * np.pi)**0.5
        m_v[2] += np.sum(Delta_ij * Dz_ij) * (4 * np.pi)**0.5

    return m_v.real, m_av


def get_parity_eigenvalues(calc, ik=0, spin_orbit=False, bands=None, Nv=None,
                           inversion_center=[0, 0, 0], deg_tol=1.0e-6,
                           tol=1.0e-6):
    """Calculates parity eigenvalues at time-reversal invariant k-points.
    Only works in plane wave mode.
    """

    assert len(calc.get_bz_k_points()) == len(calc.get_ibz_k_points())

    kpt_c = calc.get_ibz_k_points()[ik]
    if Nv is None:
        Nv = int(calc.get_number_of_electrons() / 2)

    if bands is None:
        bands = range(calc.get_number_of_bands())

    # Find degenerate subspaces
    eig_n = calc.get_eigenvalues(kpt=ik)[bands]
    e_in = []
    used_n = []
    for n1, e1 in enumerate(eig_n):
        if n1 not in used_n:
            n_n = []
            for n2, e2 in enumerate(eig_n):
                if np.abs(e1 - e2) < deg_tol:
                    n_n.append(n2)
                    used_n.append(n2)
            e_in.append(n_n)

    print()
    print(f' Inversion center at: {inversion_center}')
    print(f' Calculating inversion eigenvalues at k = {kpt_c}')
    print()

    center_v = np.array(inversion_center) / Bohr
    G_Gv = calc.wfs.pd.get_reciprocal_vectors(q=ik, add_q=True)

    psit_nG = np.array([calc.wfs.kpt_u[ik].psit_nG[n]
                        for n in bands])
    if spin_orbit:
        n1 = bands[0]
        n2 = bands[-1] + 1
        assert (bands == np.arange(n1, n2)).all()
        soc = soc_eigenstates(calc, n1=n1, n2=n2)
        v_kmn = soc.eigenvectors()
        psit0_mG = np.dot(v_kmn[ik, :, ::2], psit_nG)
        psit1_mG = np.dot(v_kmn[ik, :, 1::2], psit_nG)
    for n in range(len(bands)):
        psit_nG[n] /= (np.sum(np.abs(psit_nG[n])**2))**0.5
    if spin_orbit:
        for n in range(2 * len(bands)):
            A = np.sum(np.abs(psit0_mG[n])**2) + np.sum(np.abs(psit1_mG[n])**2)
            psit0_mG[n] /= A**0.5
            psit1_mG[n] /= A**0.5

    P_GG = np.ones((len(G_Gv), len(G_Gv)), float)
    for iG, G_v in enumerate(G_Gv):
        P_GG[iG] -= ((G_Gv[:] + G_v).round(6)).any(axis=1)
    assert (P_GG == P_GG.T).all()

    phase_G = np.exp(-2.0j * np.dot(G_Gv, center_v))

    p_n = []
    print('n   P_n')
    for n_n in e_in:
        if spin_orbit:
            # The dimension of parity matrix is doubled with spinorbit
            m_m = [2 * n_n[0] + i for i in range(2 * len(n_n))]
            Ppsit0_mG = np.dot(P_GG, psit0_mG[m_m].T).T
            Ppsit0_mG[:] *= phase_G
            Ppsit1_mG = np.dot(P_GG, psit1_mG[m_m].T).T
            Ppsit1_mG[:] *= phase_G
            P_nn = np.dot(psit0_mG[m_m].conj(), np.array(Ppsit0_mG).T)
            P_nn += np.dot(psit1_mG[m_m].conj(), np.array(Ppsit1_mG).T)
        else:
            Ppsit_nG = np.dot(P_GG, psit_nG[n_n].T).T
            Ppsit_nG[:] *= phase_G
            P_nn = np.dot(psit_nG[n_n].conj(), np.array(Ppsit_nG).T)
        P_eig = np.linalg.eigh(P_nn)[0]
        if np.allclose(np.abs(P_eig), 1, tol):
            P_n = np.sign(P_eig).astype(int).tolist()
            if spin_orbit:
                # Only include one of the degenerate pair of eigenvalues
                Pm = np.sign(P_eig).tolist().count(-1)
                Pp = np.sign(P_eig).tolist().count(1)
                P_n = Pm // 2 * [-1] + Pp // 2 * [1]
            print(f'{str(n_n)[1:-1]}: {str(P_n)[1:-1]}')
            p_n += P_n
        else:
            print(f'  {n_n} are not parity eigenstates')
            print(f'     P_n: {P_eig}')
            print(f'     e_n: {eig_n[n_n]}')
            p_n += [0 for n in n_n]

    return np.ravel(p_n)


def get_L_vlmm():
    if len(_L_vlmm) == 3:
        return _L_vlmm

    s = np.array([[0.0]])
    p = np.zeros((3, 3), complex)  # y, z, x
    p[0, 1] = -1.0j
    p[1, 0] = 1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 3] = -1.0j
    d[3, 0] = 1.0j
    d[1, 2] = -3**0.5 * 1.0j
    d[2, 1] = 3**0.5 * 1.0j
    d[1, 4] = -1.0j
    d[4, 1] = 1.0j
    _L_vlmm.append([s, p, d])

    p = np.zeros((3, 3), complex)  # y, z, x
    p[1, 2] = -1.0j
    p[2, 1] = 1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 1] = 1.0j
    d[1, 0] = -1.0j
    d[2, 3] = -3**0.5 * 1.0j
    d[3, 2] = 3**0.5 * 1.0j
    d[3, 4] = -1.0j
    d[4, 3] = 1.0j
    _L_vlmm.append([s, p, d])

    p = np.zeros((3, 3), complex)  # y, z, x
    p[0, 2] = 1.0j
    p[2, 0] = -1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 4] = 2.0j
    d[4, 0] = -2.0j
    d[1, 3] = 1.0j
    d[3, 1] = -1.0j
    _L_vlmm.append([s, p, d])

    return _L_vlmm
